# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`messages` -- Message Utilities
==========================================
"""
import logging

from .. import conf, constants, exceptions
from . import random_characters

logger = logging.getLogger(__name__)


def parse_router_message(message):
    """
    Parses the generic format of an eMQP/1.0 message and returns the
    parts.

    Args:
        message: the message you wish to have parsed

    Returns (tuple) (sender_id, command, message_id, (message_body, and_data))
    """
    try:
        sender = message[0]
        # noop = message[1]
        # protocol_version = message[2]
        command = message[3]
        msgid = message[4]
    except IndexError:
        raise exceptions.InvalidMessageError('Invalid Message Encountered: %s'
                                             % str(message))
    if len(message) > 5:
        msg = message[5:]
    else:
        msg = ()
    return (sender, command, msgid, msg)


def parse_message(message):
    """
    Parses the generic format of an eMQP/1.0 message and returns the
    parts.

    Args:
        message: the message you wish to have parsed

    Returns (tuple) (command, message_id, (message_body, and_data))
    """
    try:
        # noop = message[0]
        # protocol_version = message[1]
        command = message[2]
        msgid = message[3]
    except IndexError:
        raise exceptions.InvalidMessageError('Invalid Message Encountered: %s'
                                             % str(message))
    if len(message) > 4:
        msg = message[4:]
    else:
        msg = ()
    return (command, msgid, msg)


def generate_msgid(prefix=None):
    """
    Returns a random string to be used for message ids. Optionally the ID can
    be prefixed with `prefix`.

    Args:
        prefix (str): Value to prefix on to the random part of the id. Useful
            for prefixing some meta data to use for things
    """
    id = random_characters()
    return id if not prefix else str(prefix) + id


def send_emqp_message(socket, command, message=None):
    """
    Formats and sends an eMQP message

    Args:
        socket
        command
        message
    Raises:

    Returns:
        str: Message id for this message
    """
    msgid = generate_msgid()
    msg = (str(command).upper(), msgid)
    if message and isinstance(message, (tuple, list)):
        msg += tuple(message)
    elif message:
        msg += (message,)

    socket.send_multipart(msg, constants.PROTOCOL_VERSION)

    return msgid


def send_emqp_router_message(socket, recipient_id, command, message=None):
    """
    Formats and sends an eMQP message taking into account the recipient frame
    used by a :attr:`zmq.ROUTER` device.

    Args:
        socket: socket to send the message with
        recipient_id (str): the id of the connected device to reply to
        command (str): the eMQP command to send
        message: a msg tuple to send

    Returns
        str: Message id for this message
    """
    msgid = generate_msgid()
    msg = (str(command).upper(), msgid)
    if message and isinstance(message, (tuple, list)):
        msg += message
    elif message:
        msg += (message,)

    socket.send_multipart(msg, constants.PROTOCOL_VERSION,
                          _recipient_id=recipient_id)

    return msgid


def fwd_emqp_router_message(socket, recipient_id, payload):
    """
    Forwards `payload` to socket untouched.

    .. note:
       Because it's untouched, and because this function targets
       :prop:`zmq.ROUTER`, it may be a good idea to first strip off the
       leading sender id before forwarding it. If you dont you will need to
       account for that on the recipient side.

    Args:
        socket: socket to send the message with
        recipient_id (str): the id of the connected device to reply to
        payload (tuple): The message to send. The first frame should be an
            empty string
    """
    import zmq
    import errno

    errnos = [errno.EHOSTUNREACH,
              errno.EAGAIN]

    payload = [recipient_id, ] + payload
    if conf.SUPER_DEBUG:
        logger.debug('Forwarding message: {}'.format(str(payload)))
    try:
        socket.zsocket.send_multipart(payload,
                                      flags=zmq.NOBLOCK)
    except zmq.error.ZMQError as e:
        if e.errno in errnos:
            e.message = e.message + " {}".format(recipient_id)
            raise exceptions.PeerGoneAwayError(e)
        else:
            raise exceptions.EventMQError("errno {}: {}".format(e.errno,
                                                                str(e)))



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
import time
import uuid

import zmq

from .. import conf, constants
from ..utils.classes import ZMQSendMixin, ZMQReceiveMixin
from ..utils.devices import generate_device_name


class FakeDevice(ZMQReceiveMixin, ZMQSendMixin):
    """
    A fake router device so we can test with some of the nice utilities, but
    still allowing manual control
    """
    def __init__(self, type_=zmq.ROUTER):
        super(FakeDevice, self).__init__()

        self.name = generate_device_name()
        self.zsocket = zmq.Context.instance().socket(type_)
        self.zsocket.setsockopt(zmq.IDENTITY, self.name)


def send_raw_INFORM(sock, type_, queues=(conf.DEFAULT_QUEUE_NAME,)):
    """
    Send an INFORM message to a raw ZMQ socket.

    Args:
        sock: Socket to send the message on
        type_: Type of client. One of ``scheduler`` or ``worker``
        queues (list, tuple): List of queues to listen on

    Return:
        str: The message id of the INFORM message
    """
    msgid = str(uuid.uuid4())
    tracker = sock.zsocket.send_multipart((
        '',
        constants.PROTOCOL_VERSION,
        'INFORM',
        msgid,
        ','.join(queues),
        type_
    ), copy=False, track=True)
    tracker.wait(1)

    return msgid


def send_raw_READY(sock):
    """
    Sends a READY message to a raw ZMQ socket.

    Args:
        sock: Socket to send the message on

    Return:

    """
    msgid = str(uuid.uuid4())
    tracker = sock.zsocket.send_multipart((
        '',
        constants.PROTOCOL_VERSION,
        'READY',
        msgid
    ), copy=False, track=True)
    tracker.wait(1)

    return msgid



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`jobs` -- Client Job Helpers
=================================
"""
import logging
import os

from ..constants import ENV_BROKER_ADDR
from ..sender import Sender
from . import messages

logger = logging.getLogger(__name__)


class Job(object):
    """
    Defines a deferred EventMQ job.

    Usage:

    .. code-block:: python

       from eventmq import job

       @job(queue='messaging')
       def send_email(recipient, subject, message):
           from email.mime.text import MIMEText
           import smtplib

           msg = MIMEText(message)
           msg['Subject'] = subject
           msg['From'] = 'no-reply@foobar.io'
           msg['To'] = recipient

           s = smtplib.SMTP('smtp.gmail.com')

           s.login('me@gmail.com', 'my-app-password')

           s.sendmail('me@gmail.com', [recipient,], msg.as_string())
           s.quit()
    """
    def __init__(self, broker_addr=None, queue=None, async=True, *args,
                 **kwargs):
        """
        Args:
            queue (str): Name of the queue this function should be executed
                in. If no queue provided ``default`` is used.
            broker_addr (str): Address of the broker to send the job to. If no
                address is given then the value of the environment variable
                ``EMQ_BROKER_ADDR`` will be used, If that is undefined a
                warning will be emitted and the job will be run synchronously.
            async (bool): If you want to run all executions of a particular job
                synchronously but still decorate it with the job decorator you
                can set this to False. This is useful for unit tests.

        """
        self.broker_addr = broker_addr or os.environ.get(ENV_BROKER_ADDR)
        self.queue = queue
        self.async = async

    def __call__(self, f):
        def delay(*args, **kwargs):
            if self.async and self.broker_addr:
                socket = Sender()
                socket.connect(addr=self.broker_addr)

                msgid = messages.defer_job(
                    socket, f, args=args, kwargs=kwargs, queue=self.queue)

                return msgid
            else:
                if self.async and not self.broker_addr:
                    logger.warning('No EMQ_BROKER_ADDR defined. Running '
                                   'function `{}` synchronously'.format(
                                       f.__name__))
                return f(*args, **kwargs)
        f.delay = delay

        return f


def job(func, broker_addr=None, queue=None, async=True, *args, **kwargs):
    """
    Functional decorator helper for creating a deferred eventmq job. See
    :class:`Job` for more information.
    """
    decorator = Job(queue=queue, broker_addr=broker_addr, async=async)

    if callable(func):
        return decorator(func)
    else:
        return decorator



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`exceptions` -- Exceptions
===============================
These are the exceptions that are raised by EventMQ. All exceptions should
be subclasses of :class:`EventMQError`
"""


class EventMQError(Exception):
    """
    All exceptions raised by EventMQ inherit from this base exception.
    """


class MessageError(EventMQError):
    """
    Raised when there is a problem with the structure of the message.
    """


class InvalidMessageError(MessageError):
    """
    Raise when EventMQ encounters a malformed message is encountered.
    """


class PeerGoneAwayError(EventMQError):
    """
    Raised when attempting to contact a peer that no longer exists (i.e. when
    sending a message to it)
    """


class NoAvailableWorkerSlotsError(EventMQError):
    """
    Raised when there is no available workers for a job manager.
    """


class UnknownQueueError(EventMQError):
    """
    Raised when a queue is not found in the internal list of queues.
    """


class CallableFromPathError(EventMQError):
    """
    Raised when construction of a callable from a path and callable_name fails.
    """
    pass



import json
import hashlib
import importlib
import inspect

from .. import log
from ..exceptions import CallableFromPathError

logger = log.setup_logger(__name__)


class IgnoreJSONEncoder(json.JSONEncoder):
    """JSON Encoder that ignores unknown keys."""
    def default(self, obj):
        try:
            return super(IgnoreJSONEncoder, self).default(obj)
        except TypeError:
            return None


def run_function(callable_name,
                 class_args=(), class_kwargs=None, args=(), kwargs=None):
    """Constructs a callable from `path` and `callable_name` and calls it.

    Args:
        callable_name (str): The name and path of the function you wish to run,
            eg ``eventmq.utils.functions.run_function``.
        class_args (list): If the callable is a class method, these args will
            be used to initialize the class.
        class_kwargs (dict): If the callable is a class method, these kwargs
            will be used to initialize the class.
        args (list): These arguments will be passed as parameters to the
            callable
        kwargs (dict): These keyword arguments will be passed as parameters to
            the callable.

    Return:
        object: Whatever is returned by calling the callable will be returned
            from this function.
    """
    if not class_args:
        class_args = ()

    if not class_kwargs:
        class_kwargs = {}

    if not args:
        args = ()

    if not kwargs:
        kwargs = {}

    try:
        callable_ = callable_from_name(
            callable_name, *class_args, **class_kwargs)
    except CallableFromPathError as e:
        logger.exception('Error importing callable {}: {}'.format(
            callable_name, str(e)))
        return

    return callable_(*args, **kwargs)


def arguments_hash(*args, **kwargs):
    """Takes `args` and `kwargs` and creates a unique identifier."""
    args = {
        'args': args,
        'kwargs': kwargs,
    }

    data = json.dumps(args, cls=IgnoreJSONEncoder)
    return hashlib.sha1(data).hexdigest()


def name_from_callable(func):
    """
    Builds the module path in string format for a callable.

    .. note:
       To use a callable Object, pass Class.__call__ as func and provide any
       class_args/class_kwargs. This is so side effects from pickling won't
       occur.

    Args:
        func (callable): The function or method to build the path for

    Returns:
        str: The callable path and name, eg
            ``"eventmq.utils.functions.name_from_callable"`` would be returned
            if you called this function on itself.
    """
    callable_name = None

    path = None
    # Methods also have the func_name property
    if inspect.ismethod(func):
        path = ("{}:{}".format(func.__module__, func.im_class.__name__))
        try:
            callable_name = func.func_name
        except AttributeError:
            callable_name = func.__name__
    elif inspect.isfunction(func):
        path = func.__module__
        try:
            callable_name = func.func_name
        except AttributeError:
            callable_name = func.__name__
    else:
        # We should account for another callable type so log information
        # about it
        if hasattr(func, '__class__') and isinstance(func, func.__class__):
            func_type = 'instanceobject'
        else:
            func_type = type(func)

        logger.error('Encountered unknown callable ({}) type {}'.format(
            func,
            func_type
        ))
        return None

    if not callable_name:
        logger.error(
            'Encountered callable with no name in {}'.format(func.__module__))
        return None

    if not path:
        try:
            func_name = func.func_name
        except AttributeError:
            func_name = func.__name__
        logger.error(
            'Encountered callable with no __module__ path {}'.format(
                func_name))
        return None

    return '{}.{}'.format(path, callable_name)


def split_callable_name(callable_name):
    """Split a callable name into it's path and name components.

    Args:
        callable_name (str): The callable name, with path, eg,
            ``eventmq.utils.functions.callable_from_name``.

    Returns:
        (str, str): The path and callable name, or None, None if a path cannot
            be found.
    """
    if not callable_name or '.' not in callable_name:
        return None, None

    elements = callable_name.split('.')
    path = '.'.join(elements[:-1])
    return path, elements[-1]


def callable_from_name(callable_name, *args, **kwargs):
    """Build a callable from a path and callable_name.

    This function is the opposite of `name_from_callable`.  It takes what is
    returned from `build_module_name` and converts it back to the original
    caller.

    Args:
        callable_name (str): The name (and path) of the function, eg,
            ``eventmq.utils.functions.callable_from_name``.
        *args (list): if `callable_name` is a method on a class, these
            arguments will be passed to the constructor when instantiating the
            class.
        *kwargs (dict): if `callable_name` is a method on a class, these
            arguments will be passed to the constructor when instantiating the
            class.

    Returns:
        function: The callable
    """
    path, callable_name = split_callable_name(callable_name)
    if ':' in path:
        _pksplit = path.split(':')
        s_package = _pksplit[0]
        s_cls = _pksplit[1]
    else:
        s_package = path
        s_cls = None

    try:
        package = importlib.import_module(s_package)
        reload(package)
    except Exception as e:
        raise CallableFromPathError(str(e))

    if s_cls:
        cls = getattr(package, s_cls)
        obj = cls(*args, **kwargs)
    else:
        obj = package

    try:
        callable_ = getattr(obj, callable_name)
    except AttributeError as e:
        raise CallableFromPathError(str(e))

    return callable_


def get_timeout_from_headers(headers):
    """Return the timeout value if it exists in the given headers

    Retruns:
        timeout(int): The timeout if found, else None
    """
    timeout = None
    for header in headers.split(','):
        if 'timeout:' in header:
            timeout = int(header.split(':')[1])
    return timeout



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
log module for eventmq

this needs so much work.
"""
import logging

import zmq
import zmq.log.handlers


FORMAT_STANDARD = logging.Formatter(
    '%(asctime)s - %(name)s  %(levelname)s - %(message)s')
FORMAT_NAMELESS = logging.Formatter(
    '%(asctime)s - %(levelname)s - %(message)s')


class PUBHandler(zmq.log.handlers.PUBHandler):
    """
    """
    pass


class handlers(object):
    """
    log handlers

    PUBLISH_HANDLER - blast logs through a pub mechanism
    STREAM_LOGGER - logs to stdout/stderr
    """
    PUBLISH_HANDLER = PUBHandler
    STREAM_HANDLER = logging.StreamHandler


def setup_logger(base_name, formatter=FORMAT_STANDARD,
                 handler=handlers.STREAM_HANDLER):

    logger = logging.getLogger(base_name)
    logger.setLevel(logging.DEBUG)

    # remove handlers we don't want
    #for h in logger.handlers:
    #    logger.removeHandler(h)

    if handler == handlers.PUBLISH_HANDLER:
        _handler_sock = zmq.Context.instance().socket(zmq.PUB)
        _handler_sock.bind('tcp://127.0.0.1:33445')

        import time
        time.sleep(1)

        handler = handler(_handler_sock)
        handler.root_topic = base_name
    else:
        handler = handler()

    handler.setFormatter(formatter)

    logger.addHandler(handler)

    return logger



class STATUS(object):
    wtf = -1          # Something went wrong
    ready = 100       # Waiting to connect or listen
    starting = 101    # Starting to bind
    listening = 102   # bound
    connecting = 200
    connected = 201
    stopping = 300
    stopped = 301


class CLIENT_TYPE(object):
    worker = 'worker'
    scheduler = 'scheduler'


# See doc/protocol.rst
PROTOCOL_VERSION = 'eMQP/1.0'

# PROTOCOL COMMANDS
DISCONNECT = "DISCONNECT"
KBYE = "KBYE"

# ADMINISTRATIVE COMMANDS
ROUTER_SHOW_WORKERS = 'ROUTER_SHOW_WORKERS'
ROUTER_SHOW_SCHEDULERS = 'ROUTER_SHOW_SCHEDULERS'

# ENVIRONMENT VARIABLES
ENV_BROKER_ADDR = 'EMQ_BROKER_ADDR'



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`poller` -- Poller
=======================
Device for polling sockets
"""
import logging
import uuid

import zmq
from zmq import Poller as ZPoller


logger = logging.getLogger(__name__)

POLLIN = zmq.POLLIN
POLLOUT = zmq.POLLOUT
POLLINOUT = zmq.POLLIN | zmq.POLLOUT


class Poller(ZPoller):
    """
    :class:`zmq.Poller` based class.
    """
    def __init__(self, *args, **kwargs):
        """
        See :class:`zmq.Poller`
        """
        super(Poller, self).__init__(*args, **kwargs)
        self.name = str(uuid.uuid4())

        self._sockets = []

    def register(self, socket, flag=0):
        """
        Register a socket to be polled by this poller

        Args:
            socket (socket): The socket object to register
            flags (int): :attr:`POLLIN`, :attr:`POLLOUT`, or
                :attr:`POLLIN`|:attr:`POLLOUT` for both. If undefined the
                socket remains unregistered.
        """
        if not flag:
            logger.warning("Leaving socket %s unregistered because no flag "
                           "was defined" % socket.name)
            return
        logger.debug('Registering %s with poller. (flags:%s)' % (socket.name,
                                                                 flag))
        self._sockets.append(socket)

        super(Poller, self).register(socket.zsocket, flag)

    def unregister(self, socket):
        """
        Unregister a socket from being polled

        Args:
            socket (socket): The socket object to registering
        """
        logger.debug("Unregistering %s from poller" % socket.name)
        if socket not in self._sockets:
            logger.warning("Attempt to unregister unregistered socket from "
                           "poller: socket: %s" % socket.name)

        self._sockets.remove(socket)
        super(Poller, self).unregister(socket.zsocket)

    def poll(self, timeout=1):
        """
        Calling :meth:`zmq.Poller.poll` directly returns a tuple set. This
        method typecasts to a dictionary for convience using the socket object
        as the key. If a socket doesn't appear in the returned dictionary then
        no events happened.

        Args:
            timeout (int): How long should poller wait for before iterating
                the next loop.

        Returns (dict) Dictionary using the socket as the key and the event the
            socket generated as the value
        """
        events = dict(super(Poller, self).poll(timeout))
        return_events = {}

        for socket in self._sockets:
            if socket.zsocket not in events:
                continue

            return_events[socket] = events[socket.zsocket]

        return return_events



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`scheduler` -- Scheduler
=============================
Handles cron and other scheduled tasks
"""
import json
import logging
import redis

from croniter import croniter
from six import next

from . import conf, constants
from constants import KBYE
from .sender import Sender
from .poller import Poller, POLLIN
from .utils.classes import EMQPService, HeartbeatMixin
from json import loads as deserialize
from json import dumps as serialize
from .utils.messages import send_emqp_message as sendmsg
from .utils.settings import import_settings
from .utils.timeutils import IntervalIter
from .utils.timeutils import seconds_until, timestamp, monotonic
from .client.messages import send_request

from eventmq.log import setup_logger

logger = logging.getLogger(__name__)
CRON_CALLER_ID = -1
INFINITE_RUN_COUNT = -1


class Scheduler(HeartbeatMixin, EMQPService):
    """
    Keeper of time, master of schedules
    """
    SERVICE_TYPE = constants.CLIENT_TYPE.scheduler

    def __init__(self, *args, **kwargs):
        self.name = kwargs.get('name', None)

        logger.info('Initializing Scheduler...')
        import_settings()
        super(Scheduler, self).__init__(*args, **kwargs)
        self.outgoing = Sender()
        self._redis_server = None

        # contains dict of 4-item lists representing cron jobs
        # key of this dictionary is a hash of caller_id, path, and callable
        # from the message of the SCHEDULE command received
        # IDX     Description
        # 0 = the next ts this job should be executed in
        # 1 = the function to be executed
        # 2 = the croniter iterator for this job
        # 3 = the queue to execute the job in
        self.cron_jobs = {}

        # contains dict of 5-item lists representing jobs based on an interval
        # key of this dictionary is a hash of caller_id, path, and callable
        # from the message of the SCHEDULE command received
        # values of this list follow this format:
        # IDX     Descriptions
        # 0 = the next (monotonic) ts that this job should be executed in
        # 1 = the function to be executed
        # 2 = the interval iter for this job
        # 3 = the queue to execute the job in
        # 4 = run_count: # of times to execute this job
        self.interval_jobs = {}

        self.poller = Poller()

        self.load_jobs()

        self._setup()

    def load_jobs(self):
        """
        Loads the jobs that need to be scheduled
        """
        raw_jobs = (
            # ('* * * * *', 'eventmq.scheduler.test_job'),
        )
        ts = int(timestamp())
        for job in raw_jobs:
            # Create the croniter iterator
            c = croniter(job[0])
            path = '.'.join(job[1].split('.')[:-1])
            callable_ = job[1].split('.')[-1]

            msg = ['run', {
                'path': path,
                'callable': callable_
            }]

            # Get the next time this job should be run
            c_next = next(c)
            if ts >= c_next:
                # If the next execution time has passed move the iterator to
                # the following time
                c_next = next(c)

            cron_hash = self.cron_hash(caller_id=CRON_CALLER_ID,
                                       path=path,
                                       callable_=callable_)

            self.cron_jobs[cron_hash] = [c_next, json.dumps(msg), c, None]

        try:
            interval_job_list = self.redis_server.lrange('interval_jobs',
                                                         0,
                                                         -1)
            if interval_job_list is not None:
                for i in interval_job_list:
                    logger.debug('Restoring job with hash %s' % i)
                    if (self.redis_server.get(i)):
                        self.load_job_from_redis(
                            message=deserialize(self.redis_server.get(i)))
                    else:
                        logger.warning('Expected scheduled job in redis,' +
                                       'but none was found with hash %s' % i)
        except redis.ConnectionError:
            logger.warning('Could not contact redis server')
        except Exception as e:
            logger.warning(str(e))

    def _start_event_loop(self):
        """
        Starts the actual event loop. Usually called by :meth:`Scheduler.start`
        """
        while True:
            if self.received_disconnect:
                break

            ts_now = int(timestamp())
            m_now = monotonic()
            events = self.poller.poll()

            if events.get(self.outgoing) == POLLIN:
                msg = self.outgoing.recv_multipart()
                self.process_message(msg)

            # TODO: distribute me!
            for hash_, cron in self.cron_jobs.items():
                # If the time is now, or passed
                if cron[0] <= ts_now:
                    msg = cron[1]
                    queue = cron[3]

                    # Run the msg
                    logger.debug("Time is: %s; Schedule is: %s - Running %s"
                                 % (ts_now, cron[0], msg))

                    self.send_request(msg, queue=queue)

                    # Update the next time to run
                    cron[0] = next(cron[2])
                    logger.debug("Next execution will be in %ss" %
                                 seconds_until(cron[0]))

            cancel_jobs = []
            for k, v in self.interval_jobs.iteritems():
                # The schedule time has elapsed
                if v[0] <= m_now:
                    msg = v[1]
                    queue = v[3]

                    logger.debug("Time is: %s; Schedule is: %s - Running %s"
                                 % (ts_now, v[0], msg))

                    if v[4] != INFINITE_RUN_COUNT:
                        # Decrement run_count
                        v[4] -= 1
                        # If run_count was 0, we cancel the job
                        if v[4] <= 0:
                            cancel_jobs.append(k)
                        # Otherwise we run the job
                        else:
                            # Send job and update next schedule time
                            self.send_request(msg, queue=queue)
                            v[0] = next(v[2])
                            # Rename redis key and save new run_count counter
                            try:
                                self.redis_server.rename(k,
                                                         self.schedule_hash(v))
                                self.redis_server.set(self.schedule_hash(v),
                                                      serialize(v))
                                self.redis_server.save()
                            except redis.ConnectionError:
                                logger.warning("Couldn't contact redis server")
                            except Exception as e:
                                logger.warning(
                                    'Unable to update key in redis '
                                    'server: {}'.format(e.message))
                    else:
                        # Scheduled job is in running infinitely
                        # Send job and update next schedule time
                        self.send_request(msg, queue=queue)
                        v[0] = next(v[2])
                        # Persist changes to redis
                        try:
                            self.redis_server.set(
                                self.schedule_hash(v), serialize(v))
                            self.redis_server.save()
                        except redis.ConnectionError:
                            logger.warning("Couldn't contact redis server")
                        except Exception as e:
                            logger.warning(
                                'Unable to update key in redis '
                                'server: {}'.format(e.message))

            for job in cancel_jobs:
                message = self.interval_jobs[k][1]
                self.unschedule_job(message)
                del self.interval_jobs[k]

            if not self.maybe_send_heartbeat(events):
                break

    @property
    def redis_server(self):
        # Open connection to redis server for persistance
        if self._redis_server is None:
            try:
                self._redis_server = \
                    redis.StrictRedis(host=conf.RQ_HOST,
                                      port=conf.RQ_PORT,
                                      db=conf.RQ_DB,
                                      password=conf.RQ_PASSWORD)
                return self._redis_server

            except Exception as e:
                logger.warning('Unable to connect to redis server: {}'.format(
                    e.message))
        else:
            return self._redis_server

    def send_request(self, jobmsg, queue=None):
        """
        Send a request message to the broker

        Args:
            jobmsg: The message to send to the broker
            queue: The name of the queue to use_impersonation

        Returns:
            str: ID of the message
        """
        jobmsg = json.loads(jobmsg)
        msgid = send_request(self.outgoing, jobmsg, queue=queue)

        return msgid

    def on_disconnect(self, msgid, message):
        logger.info("Received DISCONNECT request: {}".format(message))
        self._redis_server.connection_pool.disconnect()
        sendmsg(self.outgoing, KBYE)
        self.outgoing.unbind(conf.SCHEDULER_ADDR)
        super(Scheduler, self).on_disconnect(msgid, message)

    def on_kbye(self, msgid, msg):
        if not self.is_heartbeat_enabled:
            self.reset()

    def on_unschedule(self, msgid, message):
        """
           Unschedule an existing schedule job, if it exists
        """
        logger.info("Received new UNSCHEDULE request: {}".format(message))

        # TODO: Notify router whether or not this succeeds
        self.unschedule_job(message)

    def unschedule_job(self, message):
        """
        Unschedules a job if it exists based on the message used to generate it
        """
        schedule_hash = self.schedule_hash(message)

        if schedule_hash in self.interval_jobs:
            # Remove scheduled job
            self.interval_jobs.pop(schedule_hash)
        elif schedule_hash in self.cron_jobs:
            # Remove scheduled job
            self.cron_jobs.pop(schedule_hash)
        else:
            logger.warning("Couldn't find matching schedule for unschedule " +
                           "request")

        # Double check the redis server even if we didn't find the hash
        # in memory
        try:
            if (self.redis_server.get(schedule_hash)):
                self.redis_server.lrem('interval_jobs', 0, schedule_hash)
                self.redis_server.save()
        except redis.ConnectionError:
            logger.warning('Could not contact redis server')
        except Exception as e:
            logger.warning(str(e))

    def load_job_from_redis(self, message):
        """
        """
        from .utils.timeutils import IntervalIter

        queue = message[0].encode('utf-8')
        headers = message[1]
        interval = int(message[2])
        inter_iter = IntervalIter(monotonic(), interval)
        schedule_hash = self.schedule_hash(message)
        cron = message[4] if interval == -1 else ""
        ts = int(timestamp())

        # Positive intervals are valid
        if interval >= 0:
            self.interval_jobs[schedule_hash] = [
                next(inter_iter),
                message[3],
                inter_iter,
                queue,
                self.get_run_count_from_headers(headers)
            ]
        # Non empty strings are valid
        # Expecting '* * * * *' etc.
        elif cron and cron != "":
            # Create the croniter iterator
            c = croniter(cron)

            # Get the next time this job should be run
            c_next = next(c)
            if ts >= c_next:
                # If the next execution time has passed move the iterator to
                # the following time
                c_next = next(c)

            self.cron_jobs[schedule_hash] = [c_next, message[3], c, queue]

    def on_schedule(self, msgid, message):
        """
        """
        logger.info("Received new SCHEDULE request: {}".format(message))

        queue = message[0]
        headers = message[1]
        interval = int(message[2])
        cron = str(message[4])

        schedule_hash = self.schedule_hash(message)

        # Notify if this is updating existing, or new
        if (schedule_hash in self.cron_jobs or
                schedule_hash in self.interval_jobs):
            logger.debug('Update existing scheduled job with %s'
                         % schedule_hash)
        else:
            logger.debug('Creating a new scheduled job with %s'
                         % schedule_hash)

        # If interval is negative, cron MUST be populated
        if interval >= 0:
            inter_iter = IntervalIter(monotonic(), interval)

            self.interval_jobs[schedule_hash] = [
                next(inter_iter),
                message[3],
                inter_iter,
                queue,
                self.get_run_count_from_headers(headers)
            ]

            if schedule_hash in self.cron_jobs:
                self.cron_jobs.pop(schedule_hash)
        else:
            ts = int(timestamp())
            c = croniter(cron)
            c_next = next(c)
            if ts >= c_next:
                # If the next execution time has passed move the iterator to
                # the following time
                c_next = next(c)

            self.cron_jobs[schedule_hash] = [c_next,
                                             message[3],
                                             c,
                                             None]

            if schedule_hash in self.interval_jobs:
                self.interval_jobs.pop(schedule_hash)

        # Persist the scheduled job
        try:
            if schedule_hash not in self.redis_server.lrange(
                    'interval_jobs', 0, -1):
                self.redis_server.lpush('interval_jobs', schedule_hash)
            self.redis_server.set(schedule_hash, serialize(message))
            self.redis_server.save()
        except redis.ConnectionError:
            logger.warning('Could not contact redis server')
        except Exception as e:
            logger.warning(str(e))

        if 'nohaste' not in headers:
            self.send_request(message[3], queue=queue)

    def get_run_count_from_headers(self, headers):
        run_count = INFINITE_RUN_COUNT
        for header in headers.split(','):
            if 'run_count:' in header:
                run_count = int(header.split(':')[1])
        return run_count

    def on_heartbeat(self, msgid, message):
        """
        Noop command. The logic for heartbeating is in the event loop.
        """

    def cron_hash(self, caller_id, path, callable_):
        schedule_hash_items = {'caller_id': caller_id,
                               'path': path,
                               'callable': callable_}

        # Hash the sorted, immutable set of items in our identifying dict
        schedule_hash = str(hash(tuple(frozenset(sorted(
            schedule_hash_items.items())))))

        return schedule_hash

    def schedule_hash(self, message):
        """
        Create a unique identifier for this message for storing
        and referencing later
        """
        # Items to use for uniquely identifying this scheduled job
        # TODO: Pass caller_id in a more rigid place
        msg = deserialize(message[3])[1]
        schedule_hash_items = {'caller_id': msg['class_args'][0],
                               'path': msg['path'],
                               'callable': msg['callable']}

        # Hash the sorted, immutable set of items in our identifying dict
        schedule_hash = str(hash(tuple(frozenset(sorted(
            schedule_hash_items.items())))))

        return schedule_hash

    def scheduler_main(self):
        """
        Kick off scheduler with logging and settings import
        """
        setup_logger("eventmq")
        import_settings()
        self.__init__()
        self.start(addr=conf.SCHEDULER_ADDR)


# Entry point for pip console scripts
def scheduler_main():
    s = Scheduler()
    s.scheduler_main()


def test_job():
    print "hello!"



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
import unittest

import mock
from testfixtures import LogCapture

from ..client import jobs

class TestCase(unittest.TestCase):
    BROKER_ADDR = 'tcp://127.0.0.1:3748'

    @mock.patch('eventmq.client.messages.defer_job')
    @mock.patch('eventmq.client.jobs.Sender')
    def test_job_decorator(self, Sender_mock, defer_job_mock):
        defer_job_mock.return_value = '43e14eaa-2034-4c84-8fe7-5577c70b6a7c'
        d = jobs.job(test_func, broker_addr=self.BROKER_ADDR)

        self.assertTrue(hasattr(d, 'delay'))

        self.assertEqual(d(), 12)
        self.assertEqual(d.delay(), '43e14eaa-2034-4c84-8fe7-5577c70b6a7c')
        defer_job_mock.assert_called_with(Sender_mock(), test_func, args=(),
                                          kwargs={}, queue=None)

        d = jobs.job(test_func, broker_addr=self.BROKER_ADDR, queue='mojo')
        d.delay(1, 2, three=3)
        defer_job_mock.assert_called_with(Sender_mock(), test_func,
                                          args=(1, 2), kwargs={'three': 3},
                                          queue='mojo')

    @mock.patch('eventmq.client.messages.defer_job')
    @mock.patch('eventmq.client.jobs.Sender')
    def test_job_decorator_no_broker(self, Sender_mock, defer_job_mock):
        defer_job_mock.return_value = '43e14eaa-2034-4c84-8fe7-5577c70b6a7c'
        d = jobs.job(test_func, queue='fut')

        with LogCapture() as log_checker:
            r = d.delay()

            log_checker.check(
                ('eventmq.client.jobs',
                 'WARNING',
                 'No EMQ_BROKER_ADDR defined. Running function `test_func` '
                 'synchronously'),)

        self.assertEqual(r, 12)

    def test_get_job_decorator(self):
        decorate = jobs.job(None)
        self.assertTrue(isinstance(decorate, jobs.Job))

        f = decorate(test_func)
        self.assertEqual(f(), 12)


def test_func():
    return 12



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
import time

from .. import conf, constants, worker
from nose import with_setup
from multiprocessing import Pool

ADDR = 'inproc://pour_the_rice_in_the_thing'


def setup_func():
    global pool
    global out
    pool = Pool()
    out = pool.map(job, range(1))

@with_setup(setup_func)
def test_run_with_timeout():
    payload = {
        'path': 'eventmq.tests.test_worker',
        'callable': 'job',
        'args': [2]
    }

    msgid = worker.run(payload, 12345, timeout=1)


def job(sleep_time=0):
    time.sleep(sleep_time)

    return True



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`messages` -- Client Messaging
===================================
"""
import logging
from json import dumps as serialize

from .. import conf
from ..utils.messages import send_emqp_message
from ..utils.functions import name_from_callable, split_callable_name

logger = logging.getLogger(__name__)


def schedule(socket, func, interval_secs=None, args=(), kwargs=None,
             class_args=(), class_kwargs=None, headers=('guarantee',),
             queue=conf.DEFAULT_QUEUE_NAME, unschedule=False, cron=None):
    """
    Execute a task on a defined interval.

    .. note::

       All passed class & fuction kwargs/args MUST be json serializable.

    Args:
        socket (socket): eventmq socket to use for sending the message
        func (callable): the callable (or string path to calable) to be
            scheduled on a worker
        minutes (int): minutes to wait in between executions
        args (list): list of *args to pass to the callable
        interval_secs (int): Run job every interval_secs or None if using cron
        cron (string): cron formatted string used for job schedule if
            interval_secs is None, i.e. '* * * * *' (every minute)
        kwargs (dict): dict of **kwargs to pass to the callable
        class_args (list): list of *args to pass to the class (if applicable)
        class_kwargs (dict): dict of **kwargs to pass to the class (if
            applicable)
        headers (list): list of strings denoting enabled headers. Default:
            guarantee is enabled to ensure the scheduler schedules the job.
        queue (str): name of the queue to use when executing the job. The
            default value is the default queue.
    Raises:
        TypeError: When one or more parameters are not JSON serializable.
    Returns:
       str: ID of the schedule message that was sent. None if there was an
           error
    """
    if not class_kwargs:
        class_kwargs = {}
    if not kwargs:
        kwargs = {}

    if not len(class_args) > 0 and not cron:
        logger.error('First `class_args` argument must be caller_id for '
                     'scheduling interval jobs')
        return

    if not unschedule and \
       ((interval_secs and cron) or (not interval_secs and not cron)):
        logger.error('You must sepcify either `interval_secs` or `cron`, '
                     'but not both (or neither)')
        return

    if func and isinstance(func, basestring):
        if '.' not in func:
            logger.error('Invalid callable string passed, '
                         'absolute path required: "{}"'.format(func))
            return
        path, callable_name = split_callable_name(func)
    elif callable(func):
        callable_name = name_from_callable(func)
        path, callable_name = split_callable_name(callable_name)
    else:
        logger.error('Encountered non-callable func: {}'.format(func))
        return

    if not callable_name or not path:
        logger.error('Encountered invalid callable, will not proceed.')
        return

    # TODO: convert all the times to seconds for the clock
    msg = ['run', {
        'callable': callable_name,
        'path': path,
        'args': args,
        'kwargs': kwargs,
        'class_args': class_args,
        'class_kwargs': class_kwargs,
    }]

    msgid = send_schedule_request(socket, interval_secs=interval_secs or -1,
                                  cron=cron or '',
                                  message=msg, headers=headers, queue=queue,
                                  unschedule=unschedule)

    # TODO: Return msgid only if we got some sort of ACK
    return msgid


def defer_job(
        socket, func, wrapper=None, args=(), kwargs=None, class_args=(),
        class_kwargs=None, reply_requested=False, guarantee=False,
        retry_count=0, timeout=0, debounce_secs=False,
        queue=conf.DEFAULT_QUEUE_NAME):
    """
    Used to send a job to a worker to execute via `socket`.

    This tries not to raise any exceptions so use some of the message flags to
    guarentee things.

    .. note::

       All passed class & fuction kwargs/args MUST be json serializable.

    Args:
        socket (socket): eventmq socket to use for sending the message
        func (callable or str): the callable (or string path to callable) to be
            deferred to a worker
        wrapper (callable): optional wrapper for the call to func to be
             wrapped with
        args (list): list of *args for the callable
        kwargs (dict): dict of **kwargs for the callable
        class_args (list): list of *args to pass to the the class when
            initializing (if applicable).
        class_kwargs (dict): dict of **kwargs to pass to the class when
            initializing (if applicable).
        reply_requested (bool): request the return value of func as a reply
        retry_count (int): How many times should be retried when encountering
            an Exception or some other failure before giving up. (default: 0
            or immediately fail)
        timeout (int): How many seconds should we wait before killing the job
            default: 0 which means infinite timeout
        debounce_secs (secs): Number of seconds to debounce the job.   See
            `debounce_deferred_job` for more information.
        queue (str): Name of queue to use when executing the job. If this value
            evaluates to False, the default is used. Default: is configured
            default queue name
    Raises:
        TypeError: When one or more parameters are not JSON serializable.
    Returns:
        str: ID for the message/deferred job. This value will be None if there
            was an error.
    """
    callable_name = None
    path = None

    # Just incase this was passed None
    if not queue:
        queue = conf.DEFAULT_QUEUE_NAME

    if not class_kwargs:
        class_kwargs = {}

    if not kwargs:
        kwargs = {}

    if func and isinstance(func, basestring):
        if '.' not in func:
            logger.error('Invalid callable string passed, '
                         'absolute path required: "{}"'.format(func))
            return
        path, callable_name = split_callable_name(callable_name)
    elif callable(func):
        callable_name = name_from_callable(func)
        path, callable_name = split_callable_name(callable_name)
    else:
        logger.error('Encountered non-callable func: {}'.format(func))
        return

    if wrapper and callable(wrapper):
        # Prepend the original path and callable name to args
        args = (path, callable_name, args)
        callable_name = name_from_callable(wrapper)
        path, callable_name = split_callable_name(callable_name)
    elif wrapper:
        logger.error('Encountered non-callable wrapper: {}'.format(wrapper))
        return

    if not callable_name or not path:
        logger.error('Encountered invalid callable, will not proceed.')
        return

    msg = ['run', {
        'callable': callable_name,
        'path': path,
        'args': args,
        'kwargs': kwargs,
        'class_args': class_args,
        'class_kwargs': class_kwargs,
    }]

    msgid = send_request(socket, msg,
                         reply_requested=reply_requested,
                         guarantee=guarantee,
                         retry_count=retry_count,
                         timeout=timeout,
                         queue=queue)

    return msgid


def send_request(socket, message, reply_requested=False, guarantee=False,
                 retry_count=0, timeout=0, queue=None):
    """
    Send a REQUEST command.

    Default headers are always all disabled by default. If they are included in
    the headers then they have been enabled.

    To execute a task, the message should be formatted as follows:
    {subcommand(str), {
        # dot path location where callable can be imported. If callable is a
        # method on a class, the class should always come last, and be
        # seperated with a colon. (So we know to instantiate on the receiving
        # end)
        'path': path(str),
        # function or method name to run
        'callable': callable(str),
        # Optional args for callable
        'args': (arg, arg),
        # Optional kwargs for callable
        'kwargs': {'kwarg': kwarg},
        # Optional class args, kwargs
        'class_args': (arg2, arg3),
        'class_kwargs': {'kwarg2': kwarg}

        }
    }
    Args:
        socket (socket): Socket to use when sending `message`
        message: message to send to `socket`
        reply_requested (bool): request the return value of func as a reply
        guarantee (bool): (Give your best effort) to guarantee that func is
            executed. Exceptions and things will be logged.
        retry_count (int): How many times should be retried when encountering
            an Exception or some other failure before giving up. (default: 0
            or immediatly fail)
        timeout (int): How many seconds should we wait before killing the job
            default: 0 which means infinite timeout
        queue (str): Name of queue to use when executing the job. Default: is
            configured default queue name

    Returns:
        str: ID of the message
    """
    headers = []

    if reply_requested:
        headers.append('reply-requested')

    if guarantee:
        headers.append('guarantee')

    if retry_count > 0:
        headers.append('retry-count:%d' % retry_count)

    if timeout > 0:
        headers.append('timeout:%d' % timeout)

    msgid = send_emqp_message(socket, 'REQUEST',
                              (queue or conf.DEFAULT_QUEUE_NAME,
                               ",".join(headers),
                               serialize(message)))

    return msgid


def send_schedule_request(socket, message, interval_secs=-1, headers=(),
                          queue=None, unschedule=False, cron=''):
    """
    Send a SCHEDULE or UNSCHEDULE command.

    Queues a message requesting that something happens on an
    interval for the scheduler.

    Args:
        socket (socket):
        job_schedule (str)
        message: Message to send socket.
        headers (list): List of headers for the message
        queue (str): name of queue the job should be executed in
    Returns:
        str: ID of the message
    """

    if unschedule:
        command = 'UNSCHEDULE'
    else:
        command = 'SCHEDULE'

    msgid = send_emqp_message(socket, command,
                              (queue or conf.DEFAULT_QUEUE_NAME,
                               ','.join(headers),
                               str(interval_secs),
                               serialize(message),
                               cron))

    return msgid



"""
derp publisher
"""
import eventmq

if __name__ == "__main__":
    p = eventmq.Publisher()
    p.connect('tcp://127.0.0.1:47330')

    for x in xrange(1, 99999999+1):
        p.send(bytes(x), topic='/topic1')

    p.close()



"""
derp subscriber
"""
import eventmq

if __name__ == "__main__":
    s = eventmq.Subscriber()
    s.connect('tcp://127.0.0.1:47331')
    s.subscribe('/topic1')

    while True:
        # block until something comes in. normally you'd do something with
        # this in another thread or something
        print 'r'
        print s.receive()



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
import io
import random
import unittest

import mock

from .. import constants
from .. import exceptions
from ..utils import messages
from ..utils import classes
from ..utils import settings


class SettingsTestCase(unittest.TestCase):
    settings_ini = "\n".join(
        ("[global]",
         "super_debug=TRuE",
         "frontend_addr=tcp://0.0.0.0:47291",
         "",
         "[jobmanager]",
         "super_debug=FalSe",
         'queues=[[50,"google"], [40,"pushes"], [10,"default"]]',
         "worker_addr=tcp://160.254.23.88:47290",
         "concurrent_jobs=9283",))

    @mock.patch('eventmq.utils.settings.os.path.exists')
    def test_import_settings_default(self, pathexists_mock):
        from .. import conf
        # sometimes the tests step on each other with this module. reloading
        # ensures fresh test data
        reload(conf)
        pathexists_mock.return_value = True

        # Global section
        # --------------
        with mock.patch('__builtin__.open',
                        return_value=io.BytesIO(self.settings_ini)):
            settings.import_settings()

        # Changed. Default is false
        self.assertTrue(conf.SUPER_DEBUG, True)

        # Default True
        self.assertTrue(conf.HIDE_HEARTBEAT_LOGS)

        # Default is 4
        self.assertEqual(conf.CONCURRENT_JOBS, 4)

        # Changed. Default is 127.0.0.1:47291
        self.assertEqual(conf.FRONTEND_ADDR, 'tcp://0.0.0.0:47291')

        # Default is (10, 'default')
        self.assertEqual(conf.QUEUES, [(10, conf.DEFAULT_QUEUE_NAME), ])

        # Job Manager Section
        # -------------------
        with mock.patch('__builtin__.open',
                        return_value=io.BytesIO(self.settings_ini)):
            settings.import_settings('jobmanager')

        # Changed
        self.assertFalse(conf.SUPER_DEBUG)
        # Changed
        self.assertEqual(conf.CONCURRENT_JOBS, 9283)

        # Changed
        self.assertEqual(conf.QUEUES,
                         [(50, 'google'), (40, 'pushes'), (10, 'default')])

        self.assertEqual(conf.WORKER_ADDR, 'tcp://160.254.23.88:47290')

        # Invalid section
        # ---------------
        # This shouldn't fail, and nothing should change
        with mock.patch('__builtin__.open',
                        return_value=io.BytesIO(self.settings_ini)):
            settings.import_settings('nonexistent_section')

        self.assertEqual(conf.CONCURRENT_JOBS, 9283)
        self.assertEqual(conf.QUEUES,
                         [(50, 'google'), (40, 'pushes'), (10, 'default')])
        self.assertEqual(conf.WORKER_ADDR, 'tcp://160.254.23.88:47290')


class EMQPServiceTestCase(unittest.TestCase):

    # pretend to be an emq socket
    outgoing = 'some-outgoing-socket'

    def get_worker(self):
        """return an EMQPService mimicking a worker"""
        obj = classes.EMQPService()
        obj.SERVICE_TYPE = constants.CLIENT_TYPE.worker
        obj.outgoing = self.outgoing
        obj._meta = {
            'last_sent_heartbeat': 0
        }

        return obj

    def get_scheduler(self):
        """return an EMQPService mimicking a scheduler"""
        obj = classes.EMQPService()
        obj.SERVICE_TYPE = constants.CLIENT_TYPE.scheduler
        obj.outgoing = self.outgoing
        obj._meta = {
            'last_sent_heartbeat': 0
        }

        return obj

    @mock.patch('eventmq.utils.classes.sendmsg')
    def test_send_inform_return_msgid(self, sendmsg_mock):
        obj = self.get_worker()
        sendmsg_mock.return_value = 'some-msgid'

        retval = obj.send_inform(queues=[(10, 'default'), ])

        self.assertEqual(retval, sendmsg_mock.return_value)

    @mock.patch('eventmq.utils.classes.sendmsg')
    def test_send_inform_empty_queue_name(self, sendmsg_mock):
        obj = self.get_worker()

        obj.send_inform()

        sendmsg_mock.assert_called_with(
            self.outgoing, 'INFORM',
            ['', constants.CLIENT_TYPE.worker])

    @mock.patch('eventmq.utils.classes.sendmsg')
    def test_send_inform_scheduler(self, sendmsg_mock):
        obj = self.get_scheduler()

        obj.send_inform(queues="this shouldn't matter")

        sendmsg_mock.assert_called_with(
            self.outgoing, 'INFORM',
            ['', constants.CLIENT_TYPE.scheduler]
        )

    @mock.patch('eventmq.utils.classes.sendmsg')
    def test_send_inform_specified_valid_queues(self, sendmsg_mock):
        obj = self.get_worker()

        obj.send_inform(queues=([10, 'push'], [7, 'email'],
                                [3, 'default']))
        sendmsg_mock.asert_called_with(
            'some-outgoing-socket', 'INFORM',
            ['[10, "push"],[7, "email"],[3, "default"]',
             constants.CLIENT_TYPE.worker]
        )

    @mock.patch('eventmq.utils.classes.sendmsg')
    def test_send_inform_update_last_sent_heartbeat(self, sendmsg_mock):
        obj = self.get_worker()

        obj.send_inform(queues=(['', constants.CLIENT_TYPE.worker]))

        self.assertGreater(obj._meta['last_sent_heartbeat'], 0)


class TestCase(unittest.TestCase):
    def test_generate_msgid(self):
        msgid = messages.generate_msgid()

        self.assertEqual(type(msgid), str)

    def test_parse_message(self):
        emq_headers = ('myid', '', 'protoversion', 'command', 'msgid')
        emq_frame_singlemsg = emq_headers + ('my message',)
        emq_frame_manymsg = emq_headers + ('many', 'parts')
        emq_frame_nomsg = emq_headers

        singlemsg = messages.parse_router_message(emq_frame_singlemsg)
        manymsg = messages.parse_router_message(emq_frame_manymsg)
        nomsg = messages.parse_router_message(emq_frame_nomsg)

        self.assertEqual(singlemsg[0], emq_frame_singlemsg[0])
        self.assertEqual(singlemsg[1], emq_frame_singlemsg[3])
        self.assertEqual(singlemsg[2], emq_frame_singlemsg[4])
        self.assertEqual(singlemsg[3], (emq_frame_singlemsg[5],))

        self.assertEqual(manymsg[0], emq_frame_manymsg[0])
        self.assertEqual(manymsg[1], emq_frame_manymsg[3])
        self.assertEqual(manymsg[2], emq_frame_manymsg[4])
        self.assertEqual(manymsg[3], emq_frame_manymsg[5:])

        self.assertEqual(nomsg[0], emq_frame_nomsg[0])
        self.assertEqual(nomsg[1], emq_frame_nomsg[3])
        self.assertEqual(nomsg[2], emq_frame_nomsg[4])
        self.assertEqual(nomsg[3], ())

        broken_message = ('dlkajfs', 'lkasdjf')
        with self.assertRaises(exceptions.InvalidMessageError):
            messages.parse_router_message(broken_message)

    @unittest.skip
    def test_parse_router_message(self):
        ['aef451a0-5cef-4f03-818a-221061c8ab68', '', 'eMQP/1.0', 'INFORM',
         '5caeb5fd-15d4-4b08-89e8-4e536672eef3', 'default', 'worker']

    def test_emqDeque(self):

        full = random.randint(1, 100)
        pfull = random.randint(1, full-1)

        q = classes.EMQdeque(full=full,
                             pfull=pfull)

        # P Fill
        for i in range(0, pfull):
            q.append(i)
        self.assertTrue(q.is_pfull())
        self.assertFalse(q.is_full())

        while len(q) > 0:
            q.pop()

        # Fill
        for i in range(0, full*4):
            q.append(i)

        self.assertTrue(q.is_full())
        self.assertTrue(q.is_pfull())

        # Is iterable?
        for i in q:
            assert True

        # Check overflow
        self.assertEqual(len(q), full)

        # Remove everything we tried to insert
        for i in range(0, full*4):
            if i in q:
                q.remove(i)
            elif len(q) > 0:
                q.popleft()

        # Assert empty
        self.assertEqual(len(q), 0)
        self.assertTrue(q.is_empty())

        for i in q:
            q.remove(i)

        self.assertEqual(len(q), 0)



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`sender` -- Sender
=======================
The sender is responsible for sending messages
"""
import logging
import uuid

import zmq

from . import constants, exceptions
from .utils.classes import ZMQReceiveMixin, ZMQSendMixin

logger = logging.getLogger(__name__)


class Sender(ZMQSendMixin, ZMQReceiveMixin):
    """
    Sends messages to a particular socket

    Attributes:
        name (str): Name of this socket
        zcontext (:class:`zmq.Context`): socket context
        zsocket (:class:`zmq.Socket`):
    """

    def __init__(self, *args, **kwargs):
        """
        .. note::
           All args are optional unless otherwise noted.

        Args:
            name (str): name of this socket. By default a uuid will be
                generated
            context (:class:`zmq.Context`): Context to use when building the
                socket
            socket (:class:`zmq.Socket`): Should be one of :attr:`zmq.REQ` or
                :attr:`zmq.DEALER`. By default a `DEALER` is used
            skip_zmqstream (bool): If set to true, skip creating the zmqstream
                socket

        """
        self.zcontext = kwargs.pop('context', zmq.Context.instance())

        # Set zsocket to none so we can check if it exists and close it before
        # rebuilding it later.
        self.zsocket = None

        self.name = kwargs.pop('name', str(uuid.uuid4()))

        self.rebuild(*args, **kwargs)

    def listen(self, addr=None):
        """
        start listening on `addr`

        Args:
            addr (str): Address to listen on as a connction string

        Raises:
            :class:`Exception`
        """
        if self.ready:
            self.zsocket.bind(addr)
            self.status = constants.STATUS.listening
            logger.info('Listening on %s' % (addr))
        else:
            raise exceptions.EventMQError('Not ready. status=%s' %
                                          (self.status))

    def connect(self, addr=None):
        """
        Connect to address defined by `addr`

        Args:
            addr (str): Address to connect to as a connection string

        Raises:
            :class:`Exception`
        """
        if self.ready:
            self.zsocket.connect(addr)
            self.status = constants.STATUS.connected
            logger.debug('Connecting to %s' % (addr))
        else:
            raise exceptions.EventMQError('Not ready. status=%s' %
                                          (self.status))

    def unbind(self, addr):
        """
        Unbinds current socket

        Args:
            addr (str): Address to unbind from as a string

        Raises:
           :class:`Exception`
        """

        if self.status == constants.STATUS.listening:
            self.zsocket.unbind(addr)
            self.status = constants.STATUS.ready
        else:
            raise Exception("Receiver {} is {}, but is trying to unbind".format
                            (
                                self.name,
                                self.status))

    def rebuild(self, *args, **kwargs):
        """
        Rebuilds the socket. This is useful when you need to reconnect to
        something without restarting the process. Many of these things happen
        happen during :meth:`self.__init__`, so it takes roughly the same
        parameters as :meth:`self.__init__`

        Args:
            socket (:class:`zmq.Socket`): Should be one of :attr:`zmq.REQ` or
                :attr:`zmq.DEALER`. By default a `DEALER` is used
            skip_zmqstream (bool): If set to true, skip creating the zmqstream
                socket
        """
        if self.zsocket:
            self.zsocket.close()

        self.name = kwargs.pop('name', str(uuid.uuid4()).encode('ascii'))
        self.zsocket = kwargs.pop('socket', self.zcontext.socket(zmq.DEALER))
        self.zsocket.setsockopt(zmq.IDENTITY, self.name)

        self.status = constants.STATUS.ready

    @property
    def ready(self):
        """
        Property used to check if this receiver is ready.

        Returns:
            bool: True if the receiver is ready to connect or listen, otherwise
                False
        """
        return self.status == constants.STATUS.ready



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
import unittest

import zmq
import signal
import os

from .. import sender


class TestCase(unittest.TestCase):
    def setUp(self):
        self.zcontext = zmq.Context.instance()

        self.sender = sender.Sender()

    def test_send_multipart(self):
        # Test that multipart installs the correct headers
        socket = self.zcontext.socket(zmq.ROUTER)
        socket.bind('inproc://test_send_multipart')
        self.sender.connect('inproc://test_send_multipart')

        self.sender.send_multipart(('Hello!', ), '1')

        self.assert_(socket.poll() != 0)
        msg = socket.recv_multipart()
        self.assertEqual(msg[0], self.sender.name)
        self.assertEqual(msg[1], '')
        self.assertEqual(msg[2], '1')
        self.assertEqual(msg[3], 'Hello!')

    def test_send_multipart_unicode(self):
        # Test that send_multipart handles unicode safely
        pass

    def test_listen(self):
        # test that the listen method functions correctly
        with self.assertRaises(sender.exceptions.EventMQError):
            self.sender.status = 'something else'
            self.sender.listen('ipc://emq-test_sender.ipc')

        # put the status back
        self.sender.status = sender.constants.STATUS.ready
        self.sender.listen('ipc://emq-test_sender.ipc')
        self.assertEqual(self.sender.status, sender.constants.STATUS.listening)

    def test_connect(self):
        # test that the connect method functions correctly
        with self.assertRaises(sender.exceptions.EventMQError):
            self.sender.status = 'something else'
            self.sender.connect()

        self.sender.status = sender.constants.STATUS.ready
        self.sender.connect('ipc://emq-test_sender.ipc')

    def test_disconnect(self):
        with self.assertRaises(Exception):
            self.sender.status = sender.constants.STATUS.ready
            self.sender.listen('ipc://emq-test_sender.ipc')
            self.sender.status = 'bogus'
            self.sender.unbind('ipc://emq-test_sender.ipc')

        self.sender.status = sender.constants.STATUS.ready
        self.sender.listen('ipc://emq-test_sender.ipc')
        self.sender.unbind('ipc://emq-test_sender.ipc')

    def test_rebuild(self):
        self.sender.listen('ipc://emq-test_sender.ipc')
        self.assertEqual(self.sender.status, sender.constants.STATUS.listening)

        self.sender.rebuild()
        self.assertEqual(self.sender.status, sender.constants.STATUS.ready)



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`timeutils` -- Time Utilites
=================================
"""
try:
    from time import monotonic  as _monotonic  # Python3
except ImportError:
    from monotonic import monotonic as _monotonic
from time import time as _time


def timestamp():
    """
    """
    return _time()


def monotonic():
    """
    Returns (float) seconds since boot, or something close to it. This value
    will never count down so it's useful for cases where DST would mess up
    time.time() arithmetic (e.g. heartbeats).
    """
    return _monotonic()


def seconds_until(ts):
    """
    Calculates the number of seconds until `ts` by subtracting it from
    time.time()
    """
    return ts - timestamp()


class IntervalIter(object):
    """
    represents an interval (in seconds) and it's `next()` execution time

    Usage:
        # interval of 5min using monotonic clock (assume it starts at 0 for the
        # sake of the example)
        interval = IntervalIter(monotonic, 300)
        # Py2

        interval.next()  # 300
        interval.next()  # 600

        # Py3
        next(interval)  # 300
        next(interval)  # 600
    """
    def __init__(self, start_value, interval_secs):
        """
        Args:
            start_value (numeric) - the timestamp to begin with. usually gotten
                via :func:`monotonic` or :func:`timestamp`
            interval_secs (int) - the number of seconds between intervals
        """
        self.current = start_value
        self.interval_secs = interval_secs

        # iterate the first time so the first call to .next() is interval_secs
        # + start_value
        self.__next__()

    def __iter__(self):
        return self

    def __next__(self):  # Py3
        self.current += self.interval_secs
        return self.current - self.interval_secs

    def next(self):
        return self.__next__()



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`conf` -- Settings Definitions
===================================
"""

#: SUPER_DEBUG basically enables more debugging logs. Specifically the messages
#: at different levels in the application.
#: Default: False
SUPER_DEBUG = False

#: Don't show HEARTBEAT message when debug logging is enabled
#: Default: True
HIDE_HEARTBEAT_LOGS = True

# When a queue name isn't specified use this queue name for the default. It
# would be a good idea to have a handful of workers listening on this queue
# unless you're positive that everything specifies a queue with workers.
DEFAULT_QUEUE_NAME = 'default'
DEFAULT_QUEUE_WEIGHT = 10

# Default queues for the Job Manager to listen on. The values here should match
# the values defined on the router.
QUEUES = [(DEFAULT_QUEUE_WEIGHT, DEFAULT_QUEUE_NAME), ]

# {{{Job Manager
# How long should we wait before retrying to connect to a broker?
RECONNECT_TIMEOUT = 5  # in seconds

# Don't bother with HEARTBEATS, both sending and paying attention to them
DISABLE_HEARTBEATS = False
# Assume the peer is dead after this many missed heartbeats
HEARTBEAT_LIVENESS = 3
# Assume a missed heartbeat after this many seconds
HEARTBEAT_TIMEOUT = 5
# How often should a heartbeat be sent? This should be lower than
# HEARTBEAT_TIMEOUT for the broker
HEARTBEAT_INTERVAL = 3
# Default configuration file
CONFIG_FILE = '/etc/eventmq.conf'
# Default addresses to localhost
# Router:
FRONTEND_ADDR = 'tcp://127.0.0.1:47291'
BACKEND_ADDR = 'tcp://127.0.0.1:47290'
# Where the Scheduler should connect.
SCHEDULER_ADDR = 'tcp://127.0.0.1:47291'
# Where the worker should connect
WORKER_ADDR = 'tcp://127.0.0.1:47290'
WORKER_ADDR_DEFAULT = 'tcp://127.0.0.1:47290'
WORKER_ADDR_FAILOVER = 'tcp://127.0.0.1:47290'
# Used to monitor and manage the devices
ADMINISTRATIVE_ADDR = 'tcp://127.0.0.1:47293'

# How many jobs should the job manager concurrently handle?
CONCURRENT_JOBS = 4
HWM = 10000

# Redis settings
RQ_HOST = 'localhost'
RQ_PORT = 6379
RQ_DB = 0
RQ_PASSWORD = ''

# }}}



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`settings` -- Settings Utilities
=====================================
"""
import ConfigParser
import json
import logging
import os

from . import tuplify
from .. import conf


logger = logging.getLogger(__name__)


def import_settings(section='global'):
    """
    Import settings and apply to configuration globals

    Args:
       section (str): Name of the INI section to import
    """
    config = ConfigParser.ConfigParser()

    if os.path.exists(conf.CONFIG_FILE):
        config.read(conf.CONFIG_FILE)

        if not config.has_section(section):
            logger.warning(
                'Tried to read nonexistent section {}'.format(section))
            return

        for name, value in config.items(section):
            if hasattr(conf, name.upper()):
                default_value = getattr(conf, name.upper())
                t = type(default_value)
                if isinstance(default_value, (list, tuple)):
                    try:
                        value = t(json.loads(value))
                    except ValueError:
                        raise ValueError(
                            "Invalid JSON syntax for {} setting".format(name))
                    # json.loads coverts all arrays to lists, but if the first
                    # element in the default is a tuple (like in QUEUES) then
                    # convert those elements, otherwise whatever it's type is
                    # correct
                    if isinstance(default_value[0], tuple):
                        setattr(conf, name.upper(),
                                t(map(tuplify, value)))
                    else:
                        setattr(conf, name.upper(), t(value))
                elif isinstance(default_value, bool):
                    setattr(conf, name.upper(),
                            True if 't' in value.lower() else False)
                else:
                    setattr(conf, name.upper(), t(value))
                logger.debug("Setting conf.{} to {}".format(
                    name.upper(), getattr(conf, name.upper())))
            else:
                logger.warning('Tried to set invalid setting: %s' % name)
    else:
        logger.warning('Config file at {} not found. Continuing with '
                       'defaults.'.format(conf.CONFIG_FILE))



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
import json
import unittest
import uuid

from freezegun import freeze_time
import mock
from testfixtures import LogCapture
import zmq

from eventmq import conf, constants, exceptions, receiver, router
from eventmq.utils.classes import EMQdeque
from eventmq.utils.timeutils import monotonic


class TestCase(unittest.TestCase):
    def setUp(self):
        self.router = router.Router(skip_signal=True)
        self.router.zcontext = mock.Mock(spec=zmq.Context)
        self.router.incoming = mock.Mock(spec=receiver.Receiver)
        self.router.outgoing = mock.Mock(spec=receiver.Receiver)

    @mock.patch('eventmq.receiver.zmq.Socket.bind')
    @mock.patch('eventmq.router.Router._start_event_loop')
    def test_start(self, event_loop_mock, zsocket_bind_mock):
        # Test default args
        self.router.start()
        self.router.incoming.listen.assert_called_with(conf.FRONTEND_ADDR)
        self.router.outgoing.listen.assert_called_with(conf.BACKEND_ADDR)
        self.assertEqual(self.router.status, constants.STATUS.listening)

        # Test invalid args
        # self.router.start(frontend_addr=23, backend_addr=44)

    @mock.patch('eventmq.router.Router.send_ack')
    @mock.patch('eventmq.router.Router.add_worker')
    def test_on_inform_worker(self, add_worker_mock, send_ack_mock):
        sender_id = 'omgsenderid19'
        queues = '[[32, "top"], [23, "drop"], [12, "shop"]]'
        inform_msgid = 'msg31'

        self.router.on_inform(
            sender_id, inform_msgid, [queues, 'worker'])

        self.router.send_ack.assert_called_with(
            self.router.outgoing, sender_id, inform_msgid)

        self.router.add_worker.assert_called_with(
            sender_id, [(32, 'top'), (23, 'drop'), (12, 'shop')])

    @mock.patch('eventmq.router.Router.send_ack')
    @mock.patch('eventmq.router.Router.add_worker')
    def test_on_inform_worker_default_queue(self, add_worker_mock,
                                            send_ack_mock):
        # Test on_inform when no queue is specified
        sender_id = 'omgsender18'
        queues = ''
        inform_msgid = 'msg29'

        conf.QUEUES = [(10, 'default'), ]

        self.router.on_inform(
            sender_id, inform_msgid, [queues, constants.CLIENT_TYPE.worker])

        self.router.send_ack.assert_called_with(
            self.router.outgoing, sender_id, inform_msgid)
        self.router.add_worker.assert_called_with(
            sender_id, [(10, 'default'), ])

    def test_on_inform_invalid_queues(self):
        # https://github.com/enderlabs/eventmq/issues/33
        # when receiving an invalid queue name for inform, the router shouldn't
        # crash
        sender_id = 'sch01'
        msgid = 'msg01'
        msg = ["invalid queue name", constants.CLIENT_TYPE.worker]

        with LogCapture() as log_checker:
            self.router.on_inform(sender_id, msgid, msg)

            log_checker.check(
                ('eventmq.router',
                 'ERROR',
                 'Received invalid queue names in INFORM. names:{} from:{} '
                 'type:{}'.format(msg[0], sender_id, msg[1]))
            )

    # @mock.patch('eventmq.router.Router.prioritize_queue_list')
    def test_add_worker(self):
        worker1_id = 'w1'
        worker2_id = 'w2'

        queues1 = [(10, 'top'), (9, 'drop'), (8, 'shop')]
        queues2 = [(10, 'default'), (9, 'shop'), (8, 'top')]

        self.router.add_worker(worker1_id, queues=queues1)
        self.router.add_worker(worker2_id, queues=queues2)
        # added to the list of workers
        self.assertIn(worker1_id, self.router.workers)
        self.assertIn(worker2_id, self.router.workers)
        self.assertGreater(self.router.workers[worker1_id]['hb'], 0)
        # no slots yet
        self.assertEqual(self.router.workers[worker1_id]['available_slots'], 0)

        # aware of the queues
        self.assertEqual(3, len(self.router.workers[worker1_id]['queues']))
        self.assertIn((10, 'top'), self.router.workers[worker1_id]['queues'])
        self.assertIn((9, 'drop'), self.router.workers[worker1_id]['queues'])
        self.assertIn((8, 'shop'), self.router.workers[worker1_id]['queues'])

        # Worker2
        self.assertIn((10, 'default'),
                      self.router.workers[worker2_id]['queues'])
        self.assertIn((9, 'shop'), self.router.workers[worker2_id]['queues'])
        self.assertIn((8, 'top'), self.router.workers[worker2_id]['queues'])

        self.assertIn((10, worker1_id), self.router.queues['top'])
        self.assertIn((9, worker1_id), self.router.queues['drop'])
        self.assertIn((8, worker1_id), self.router.queues['shop'])

        self.assertIn((10, worker2_id), self.router.queues['default'])
        self.assertIn((9, worker2_id), self.router.queues['shop'])
        self.assertIn((8, worker2_id), self.router.queues['top'])

    def test_add_worker_invalid_queues(self):
        with self.assertRaises(TypeError):
            self.router.add_worker('83902', 8902)

    @mock.patch('eventmq.utils.messages.generate_msgid')
    def test_send_ack(self, generate_msgid_mock):
        ack_msgid = 'msg12'
        orig_msgid = 'msg3'
        sender_id = 'sender93'

        generate_msgid_mock.return_value = ack_msgid

        self.router.send_ack(self.router.outgoing, sender_id, orig_msgid)

        # Verify that an ACK was sent for the INFORM
        self.router.outgoing.send_multipart.assert_called_with(
            ('ACK', ack_msgid, orig_msgid),
            constants.PROTOCOL_VERSION, _recipient_id=sender_id)

    def test_reset_heartbeat_counters(self):
        self.router._meta['last_sent_scheduler_heartbeat'] = 392

        self.router.reset_heartbeat_counters()
        self.assertEqual(self.router._meta['last_sent_scheduler_heartbeat'], 0)

    @mock.patch('eventmq.utils.messages.generate_msgid')
    @freeze_time("2000-01-01")
    def test_send_heartbeat(self, generate_msgid_mock):
        recipient_id = 'n2'
        msgid = 'msg9'
        ts = 946684800.0  # 2000-01-01

        generate_msgid_mock.return_value = msgid

        self.router.send_heartbeat(self.router.incoming, recipient_id)

        self.router.incoming.send_multipart.assert_called_with(
            ('HEARTBEAT', msgid, str(ts)), constants.PROTOCOL_VERSION,
            _recipient_id=recipient_id)

    @mock.patch('eventmq.router.Router.send_heartbeat')
    def test_send_worker_heartbeats(self, send_heartbeat_mock):
        # last heartbeat should start at 0
        self.assertEqual(self.router._meta['last_sent_heartbeat'], 0)
        self.router.workers = {
            'w1': {
                'queues': [(10, 'default'), ],
                'hb': 123.2,
                'available_slots': 3,
            },
            'w2': {
                'queues': [(10, 'not-default'), ],
                'hb': 123.2,
                'available_slots': 2,
            }
        }
        self.router.send_workers_heartbeats()

        # After sending the heartbeat, it should be greater than 0 (note: this
        # is very hard to mock)
        self.assertGreater(self.router._meta['last_sent_heartbeat'], 0)

        send_heartbeat_mock.assert_has_calls(
            [mock.call(self.router.outgoing, 'w1'),
             mock.call(self.router.outgoing, 'w2')], any_order=True)

    @mock.patch('eventmq.router.Router.send_heartbeat')
    def test_send_schedulers_heartbeats(self, send_hb_mock):
        scheduler_id = 's39'
        self.assertEqual(self.router._meta['last_sent_scheduler_heartbeat'], 0)

        self.router.schedulers = {
            scheduler_id: {
                'hb': 0,
            }
        }

        self.router.send_schedulers_heartbeats()

        self.assertGreater(
            self.router._meta['last_sent_scheduler_heartbeat'], 0)
        send_hb_mock.assert_called_with(self.router.incoming, scheduler_id)

    def test_on_disconnect(self):
        self.assertFalse(self.router.received_disconnect)
        self.router.on_disconnect('msg1', 'derp')
        self.assertTrue(self.router.received_disconnect)

    @mock.patch('eventmq.router.fwdmsg')
    @mock.patch('eventmq.router.Router.requeue_worker')
    def test_on_ready(self, requeue_worker_mock, fwdmsg_mock):
        worker_id = '9f9z'
        msgid = 'msg18'
        msg = []

        waiting_msg = ['', constants.PROTOCOL_VERSION, 'REQUEST', msgid,
                       'default', '', 'hello world']

        self.router.workers = {
            worker_id: {
                'queues': [(10, 'default'), ],
                'hb': 123.2,
                'available_slots': 3,
            },
        }

        self.router.waiting_messages['default'] = EMQdeque(
            initial=[waiting_msg, ])

        self.router.on_ready(worker_id, msgid, msg)

        fwdmsg_mock.assert_called_with(self.router.outgoing, worker_id,
                                       waiting_msg)

        self.router.on_ready(worker_id, msgid + 'a', msg)
        self.router.requeue_worker.assert_called_with(worker_id)

    @mock.patch('eventmq.router.fwdmsg')
    @mock.patch('eventmq.router.Router.requeue_worker')
    def test_on_ready_multpile_queues(self, requeue_worker_mock,
                                      fwdmsg_mock):
        # Test that if messages are waiting on multiple queues, they are
        # dispatched immediatly after a READY message.
        worker1_id = 'w1'
        worker2_id = 'w2'

        msgid1 = 'msg21'
        msgid2 = 'msg19'
        msgid3 = 'msg6'
        waiting_msg1 = ['', constants.PROTOCOL_VERSION, 'REQUEST', msgid1,
                        'kun', '', 'hello world']
        waiting_msg2 = ['', constants.PROTOCOL_VERSION, 'REQUEST', msgid2,
                        'kun', '', 'world hello']
        waiting_msg3 = ['', constants.PROTOCOL_VERSION, 'REQUEST', msgid3,
                        'blu', '', 'goodbye']

        self.router.workers = {
            worker1_id: {
                'queues': [(10, 'kun'), (0, 'blu')],
                'hb': 123.2,
                'available_slots': 0,
            },
            worker2_id: {
                'queues': [(10, 'blu'), (0, 'kun')],
                'hb': 123.2,
                'available_slots': 0
            }
        }

        self.router.queues = {
            'kun': [(10, worker1_id), (0, worker2_id)],
            'blu': [(10, worker2_id), (0, worker1_id)],
        }

        self.router.waiting_messages = {
            'kun': EMQdeque(initial=[waiting_msg1, waiting_msg2]),
            'blu': EMQdeque(initial=[waiting_msg3, ]),
        }

        # Forward waiting_msg1
        ready_msgid1 = 'ready23'
        self.router.on_ready(worker1_id, ready_msgid1, ['READY', ready_msgid1])
        fwdmsg_mock.assert_called_with(self.router.outgoing, worker1_id,
                                       waiting_msg1)

        # Forward waiting_msg3 -- blu is a higher priority for worker2
        ready_msgid3 = 'ready19'
        self.router.on_ready(worker2_id, ready_msgid3, ['READY', ready_msgid3])
        fwdmsg_mock.assert_called_with(self.router.outgoing, worker2_id,
                                       waiting_msg3)

        # Forward waiting_msg2
        ready_msgid2 = 'ready5'
        self.router.on_ready(worker2_id, ready_msgid2, ['READY', ready_msgid2])
        fwdmsg_mock.assert_called_with(self.router.outgoing, worker2_id,
                                       waiting_msg2)

        # There should be no keys because the code checks for their existence
        # to know if there is a waiting message
        self.assertEqual(0, len(self.router.waiting_messages.keys()))

        # No waiting messages
        self.router.on_ready(worker1_id, ready_msgid1, ['READY', ready_msgid1])
        requeue_worker_mock.assert_called_with(worker1_id)
        self.router.on_ready(worker2_id, ready_msgid2, ['READY', ready_msgid2])
        requeue_worker_mock.assert_called_with(worker2_id)

    @mock.patch('eventmq.router.Router.clean_up_dead_workers')
    @mock.patch('eventmq.router.Router.process_client_message')
    @mock.patch('eventmq.router.Router.get_available_worker')
    @mock.patch('eventmq.router.fwdmsg')
    def test_on_request(self, fwdmsg_mock, get_worker_mock,
                        process_client_msg_mock, cleanupworkers_mock):
        client_id = 'c1'
        msgid = 'msg18'
        queue = 'default'
        msg = [queue, 'hello world']
        worker_id = 'w1'

        get_worker_mock.return_value = worker_id

        self.router.workers = {
            worker_id: {
                'queues': [(10, queue)],
                'hb': 2903.34,
                'available_slots': 1,
            }
        }
        self.router.queues = {
            queue: [(10, worker_id),]
        }

        # Router accepts job for 1 available slot
        self.router.on_request(client_id, msgid, msg)
        fwdmsg_mock.assert_called_with(self.router.outgoing, worker_id,
                                       ['', constants.PROTOCOL_VERSION,
                                        'REQUEST', msgid, ] + msg)
        self.assertEqual(self.router.workers[worker_id]['available_slots'], 0)

        # Router queues message when there are no workers
        def raise_no_workers(*args, **kwargs):
            raise exceptions.NoAvailableWorkerSlotsError()
        msgid += 'a'
        get_worker_mock.side_effect = raise_no_workers
        self.router.on_request(client_id, msgid, msg)

        self.assertIn(queue, self.router.waiting_messages)
        self.assertEqual(list(self.router.waiting_messages[queue])[0],
                         ['', constants.PROTOCOL_VERSION, 'REQUEST',
                          msgid] + msg)

        # Retry when the worker peer has gone away
        def raise_peer_gone_away(*args, **kwargs):
            raise exceptions.PeerGoneAwayError()
        get_worker_mock.side_effect = None
        fwdmsg_mock.side_effect = raise_peer_gone_away
        self.router.workers[worker_id]['availabe_slots'] = 1
        msgid = msgid[:-1] + 'b'

        self.router.on_request(client_id, msgid, msg)
        process_client_msg_mock.assert_called_with(
            [client_id, '', constants.PROTOCOL_VERSION, 'REQUEST', msgid]+msg,
            depth=2)

    def test_get_available_worker(self):
        worker2_id = 'w2'
        worker3_id = 'w3'

        queue1_id = 'default'
        queue2_id = 'jimjam'

        self.router.queues = {
            queue1_id: [(10, worker3_id), (0, worker2_id)],
            queue2_id: [(10, worker2_id)],
        }

        self.router.workers = {
            worker2_id: {
                'queues': [(10, queue2_id), (0, queue1_id)],
                'available_slots': 1,
            },
            worker3_id: {
                'queues': [(10, queue1_id), ],
                'available_slots': 1,
            },
        }

        # Get the next available worker for queue2
        check1 = self.router.get_available_worker(queue_name=queue2_id)
        self.assertEqual(worker2_id, check1)

        # Get the next available worker for queue1
        check2 = self.router.get_available_worker(queue_name=queue1_id)
        self.assertEqual(worker3_id, check2)

        # Pretend worker 3 is doing something
        self.router.workers[worker3_id]['available_slots'] = 0

        # Get the next available worker for queue1
        check3 = self.router.get_available_worker(queue_name=queue1_id)
        self.assertEqual(worker2_id, check3)

    def test_get_available_worker_dont_decrement_slots(self):
        # Once upon a time get_available_worker() decremented the available
        # slots counter and the townsfolk greived
        queue1_id = 'q1'
        worker1_id = 'w1'

        self.router.queues = {
            queue1_id: [(10, worker1_id, ), ]
        }

        self.router.workers = {
            worker1_id: {
                'queues': [(10, queue1_id), ],
                'available_slots': 1,
            }
        }

        self.router.get_available_worker(queue_name=queue1_id)

        self.assertEqual(self.router.workers[worker1_id]['available_slots'], 1)

    def test_requeue_worker(self):
        worker_id = 'w1'

        self.router.workers = {
            worker_id: {
                'available_slots': 1
            }
        }

        self.router.requeue_worker(worker_id)
        self.assertEqual(self.router.workers[worker_id]['available_slots'], 2)

    def test_clean_up_dead_workers(self):
        worker1_id = 'w1'
        worker2_id = 'w2'
        worker3_id = 'w3'

        queue1_id = 'default'
        queue2_id = 'jimjam'
        nonexistent_queue1 = 'nonexistent'

        # To ensure the value was changed later because monotonic() is hard to
        # mock
        self.assertEqual(self.router._meta['last_worker_cleanup'], 0)

        conf.HEARTBEAT_TIMEOUT = 1

        self.router.queues = {
            queue1_id: [(10, worker1_id), (0, worker2_id)],
            queue2_id: [(10, worker3_id), (10, worker2_id)],
        }

        self.router.workers = {
            # 3 in the future
            worker1_id: {
                'queues': [(10, queue1_id), ],
                'hb': monotonic() + 3,
                'available_slots': 0,
            },
            # below the timeout

            worker2_id: {
                'queues': [(10, queue2_id), (0, queue1_id)],
                'hb': 0,
                'available_slots': 2,
            },
            # below the timeout and a queue missing from self.router.queues
            worker3_id: {
                'queues': [(10, queue2_id), (3, nonexistent_queue1)],
                'hb': 0,
                'available_slots': 0,
            },
        }

        self.router.clean_up_dead_workers()

        self.assertIn(worker1_id, self.router.workers)
        self.assertNotIn(worker2_id, self.router.workers)
        self.assertNotIn(worker3_id, self.router.workers)

        self.assertIn(queue1_id, self.router.queues)
        self.assertNotIn(queue2_id, self.router.queues)
        self.assertNotIn(nonexistent_queue1, self.router.queues)

    @mock.patch('eventmq.router.Router.on_inform')
    @mock.patch('eventmq.router.Router.on_request')
    @mock.patch('eventmq.router.parse_router_message')
    def test_process_client_message(self, parse_msg_mock, on_request_mock,
                                    on_inform_mock):
        sender_id = 'c4'
        msgid = 'msg11'
        msg = ('hello', 'world')

        command = 'REQUEST'
        parse_msg_mock.return_value = (sender_id, command, msgid, msg)
        self.router.process_client_message(
            (sender_id, '', constants.PROTOCOL_VERSION, command, msgid) +
            msg)
        on_request_mock.assert_called_with(sender_id, msgid, msg, depth=0)

        command = 'INFORM'
        parse_msg_mock.return_value = (sender_id, command, msgid, msg)
        self.router.process_client_message(
            (sender_id, '', constants.PROTOCOL_VERSION, command, msgid) +
            msg)
        on_inform_mock.assert_called_with(sender_id, msgid, msg)

        # command = 'SCHEDULE'
        # parse_msg_mock.return_value = (sender_id, command, msgid, msg)
        # self.router.process_client_message(
        #     (sender_id, '', constants.PROTOCOL_VERSION, command, msgid) +
        #     msg)
        # on_request_mock.assert_called_with(sender_id, msgid, msg)

        # command = 'UNSCHEDULE'
        # parse_msg_mock.return_value = (sender_id, command, msgid, msg)
        # self.router.process_client_message(
        #     (sender_id, '', constants.PROTOCOL_VERSION, command, msgid) +
        #     msg)
        # on_request_mock.assert_called_with(sender_id, msgid, msg)

    @mock.patch('eventmq.router.Router.on_inform')
    @mock.patch('eventmq.router.parse_router_message')
    def test_process_worker_message(self, parse_msg_mock, on_inform_mock):
        sender_id = 'c4'
        msgid = 'msg11'
        msg = ('hello', 'world')

        command = 'INFORM'
        parse_msg_mock.return_value = (sender_id, command, msgid, msg)
        self.router.process_worker_message(
            (sender_id, '', constants.PROTOCOL_VERSION, command, msgid) +
            msg)
        on_inform_mock.assert_called_with(sender_id, msgid, msg)

    def test_prioritize_queue_list(self):
        queue = [(0, 'd'), (10, 'b'), (0, 'e'), (10, 'a'), (0, 'c')]

        sorted1 = self.router.prioritize_queue_list(queue)
        self.assertEqual([(10, 'b'), (10, 'a'), (0, 'd'), (0, 'e'),
                          (0, 'c')], sorted1)

        pop1 = sorted1.pop(0)
        self.assertEqual(pop1, (10, 'b'))
        sorted1.append(pop1)
        # a, b, d, e, c
        sorted2 = self.router.prioritize_queue_list(sorted1)
        self.assertEqual([(10, 'a'), (10, 'b'), (0, 'd'), (0, 'e'), (0, 'c')],
                         sorted2)
        pop2 = sorted2.pop(0)  # a
        pop3 = sorted2.pop(0)  # b
        pop4 = sorted2.pop(0)  # d
        self.assertEqual(pop4, (0, 'd'))
        self.assertEqual(pop2, (10, 'a'))
        self.assertEqual(pop3, (10, 'b'))
        self.assertEqual([(0, 'e'), (0, 'c')], list(sorted2))

        sorted2.append(pop2)
        sorted2.append(pop4)
        sorted2.append(pop3)

        sorted3 = self.router.prioritize_queue_list(sorted2)

        self.assertEqual(sorted3.pop(0), (10, 'a'))
        self.assertEqual(sorted3.pop(0), (10, 'b'))
        self.assertEqual(sorted3.pop(0), (0, 'e'))
        self.assertEqual(sorted3.pop(0), (0, 'c'))
        self.assertEqual(sorted3.pop(0), (0, 'd'))

    def test_disconnect(self):
        msgid = 'msg53'
        msg = ('goodbye', 'world')
        command = constants.DISCONNECT

        queue1_id = 'default'
        queue2_id = 'scoundrel'
        workers = ('w1', 'w2', 'w3')
        schedulers = ('s1', 's2', 's3')

        self.router.queues = {
            queue1_id: [(10, workers[0]), (0, workers[1])],
            queue2_id: [(10, workers[2]), (0, workers[1])]
        }

        self.router.workers = {
            workers[0]: {
                'queues': [(10, queue1_id), ],
                'hb': monotonic() + 3,
                'available_slots': 0,
            },
            workers[1]: {
                'queues': [(0, queue1_id), (10, queue2_id)],
                'hb': monotonic() + 3,
                'available_slots': 2,
            },
            workers[2]: {
                'queues': [(10, queue2_id), ],
                'hb': monotonic() + 3,
                'available_slots': 0,
            }
        }

        self.router.schedulers = {
            schedulers[0]: {
                'hb': 2903.34,
            },
            schedulers[1]: {
                'hb': 2902.99,
            },
            schedulers[2]: {
                'hb': 2904.00,
            }
        }

        self.router.process_client_message((schedulers[2], '',
                                            constants.PROTOCOL_VERSION,
                                            command, msgid) + msg)
        for i in range(0, 3):
            self.assertNotIn(schedulers[i], self.router.schedulers,
                             "Disconnect failed to remove scheduler {} from "
                             "schedulers.".format(schedulers[i]))
            self.assertNotIn(workers[i], self.router.workers,
                             "Disconnect failed to remove worker {} from "
                             "workers.".format(workers[i]))

        self.assertTrue(self.router.received_disconnect, "Router did not "
                                                         "receive disconnect.")

    def test_handle_kbye_from_scheduler(self):
        msgid = 'msg18'
        msg = ('hello', 'world')
        command = constants.KBYE

        # Scheduler ids
        s1 = 's1'
        s2 = 's2'
        s3 = 's3'

        self.router.schedulers = {
            s1: {
                'hb': 2903.34,
            },
            s2: {
                'hb': 2902.99,
            },
            s3: {
                'hb': 2904.00,
            }
        }

        self.router.scheduler_queue = [s1, s2, s3, s1, s2, s3]
        self.router.process_client_message(
            (s1, '', constants.PROTOCOL_VERSION, command, msgid) +
            msg)
        self.assertNotIn(s1, self.router.scheduler_queue,
                         'Scheduler not '
                         'removed. {'
                         '}'.format(
                             self.router.scheduler_queue))

    def test_handle_kbye_from_worker(self):
        msgid = 'msg10'
        msg = ('hello', 'world')
        command = constants.KBYE

        queue1_id = 'default'
        queue2_id = 'scallywag'

        # Worker ids
        w1 = 'w1'
        w2 = 'w2'
        w3 = 'w3'

        self.router.queues = {
            queue1_id: [(10, w1), (0, w2)],
            queue2_id: [(10, w3), (10, w2)],
        }

        self.router.workers = {
            w1: {
                'queues': [(10, queue1_id), ],
                'hb': monotonic() + 3,
                'available_slots': 0,
            },
            w2: {
                'queues': [(0, queue1_id), (10, queue2_id)],
                'hb': monotonic() + 3,
                'available_slots': 2,
            },
            w3: {
                'queues': [(10, queue2_id), ],
                'hb': monotonic() + 3,
                'available_slots': 0,
            }
        }

        self.router.process_worker_message((w2, '', constants.PROTOCOL_VERSION,
                                         command, msgid) + msg)
        self.assertNotIn(
            w2, self.router.queues[queue1_id],
            "Worker not removed from {}".format(queue1_id))
        self.assertNotIn(
            w2, self.router.queues[queue2_id],
            "Worker not removed from {}".format(queue2_id))

    def test_router_status(self):
        msgid1 = 'msg21'
        msgid2 = 'msg19'
        msgid3 = 'msg6'

        queue1_id = 'default'
        queue2_id = 'scallywag'

        # Worker ids
        w1 = 'w1'
        w2 = 'w2'
        w3 = 'w3'

        waiting_msg1 = ['', constants.PROTOCOL_VERSION, 'REQUEST', msgid1,
                        'kun', '', 'hello world']
        waiting_msg2 = ['', constants.PROTOCOL_VERSION, 'REQUEST', msgid2,
                        'kun', '', 'world hello']
        waiting_msg3 = ['', constants.PROTOCOL_VERSION, 'REQUEST', msgid3,
                        'blu', '', 'goodbye']

        self.router.queues = {
            queue1_id: [(10, w1), (0, w2)],
            queue2_id: [(10, w3), (10, w2)],
        }

        t = monotonic()

        self.router.workers = {
            w1: {
                'queues': [(10, queue1_id), ],
                'hb': t,
                'available_slots': 0,
            },
            w2: {
                'queues': [(0, queue1_id), (10, queue2_id)],
                'hb': t,
                'available_slots': 2,
            },
            w3: {
                'queues': [(10, queue2_id), ],
                'hb': t,
                'available_slots': 0,
            }
        }

        self.router.waiting_messages = {
            'kun': EMQdeque(initial=[waiting_msg1, waiting_msg2]),
            'blu': EMQdeque(initial=[waiting_msg3, ]),
        }

        # hacky, but the serialize/deserialize converts the keys to unicode
        # correctly and what not.
        self.assertEqual(
            json.loads(json.dumps({
                'job_latencies': self.router.job_latencies,
                'executed_functions': self.router.executed_functions,
                'waiting_message_counts': [
                    '{}: {}'.format(q,
                                    len(self.router.waiting_messages[q]))
                                    for q in self.router.waiting_messages]
            })),
            json.loads(self.router.get_status()))

        self.assertEqual(
            json.loads(json.dumps({
                'connected_workers': self.router.workers,
                'connected_queues': self.router.queues
            })),
            json.loads(self.router.get_workers_status()))

        self.assertEqual(
            json.loads(json.dumps({
                'connected_schedulers': self.router.schedulers
            })),
            json.loads(self.router.get_schedulers_status()))



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
import time
import unittest

import mock

from .. import conf, constants, scheduler

ADDR = 'inproc://pour_the_rice_in_the_thing'


class TestCase(unittest.TestCase):
    def test__setup(self):
        sched = scheduler.Scheduler(name='RuckusBringer')
        self.assertEqual(sched.name, 'RuckusBringer')

        self.assertFalse(sched.awaiting_startup_ack)
        self.assertEqual(sched.status, constants.STATUS.ready)

# EMQP Tests
    def test_reset(self):
        sched = scheduler.Scheduler()

        self.assertFalse(sched.awaiting_startup_ack)
        self.assertEqual(sched.status, constants.STATUS.ready)



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
import unittest

import zmq
import signal
import os

from .. import router
from .. import receiver
from .. import sender


class TestCase(unittest.TestCase):
    def setUp(self):
        self.zcontext = zmq.Context.instance()

        self.router = router.Router()
        self.receiver = self.router.incoming
        self.sender = sender.Sender()

    def test_send_multipart_unicode(self):
        # Test that send_multipart handles unicode safely
        pass

    def test_disconnect(self):
        with self.assertRaises(Exception):
            self.receiver.status = receiver.constants.STATUS.ready
            self.receiver.listen('ipc://emq-test_receiver.ipc')
            self.receiver.status = 'bogus'
            self.router.sighup_handler()

        self.receiver.status = receiver.constants.STATUS.ready
        self.receiver.listen('ipc://emq-test_receiver.ipc')
        self.receiver.unbind('ipc://emq-test_receiver.ipc')



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`worker` -- Worker Classes
===============================
Defines different short-lived workers that execute jobs
"""
from .utils.functions import run_function

from multiprocessing import Pool, TimeoutError

# the run function is executed in a different process, so we need to set the
# logger up.
from . import log

logger = log.setup_logger(__name__)


def run(payload, msgid, timeout=None):
    """
    executes job in a thread, killing it after a specified timeout
    """
    if timeout:
        worker = Pool(1)
        result = worker.apply_async(_run, args=(payload, msgid))

        try:
            out = result.get(timeout)
            return out
        except TimeoutError:
            worker.terminate()
            return (msgid, 'TimeoutError')
    else:
        return _run(payload, msgid)


def _run(payload, msgid):
    """
    process a run message and execute a job

    This is designed to run in a seperate process.
    """
    # deconstruct the payload
    path = payload.get('path')
    callable_name = payload.get('callable')
    class_args = payload.get('class_args', tuple()) or tuple()
    class_kwargs = payload.get('class_kwargs', dict()) or dict()
    args = payload.get('args', tuple()) or tuple()
    kwargs = payload.get('kwargs', dict()) or dict()

    try:
        r = run_function(
            callable_name='{}.{}'.format(path, callable_name),
            class_args=class_args or (),
            class_kwargs=class_kwargs or {},
            args=args or (),
            kwargs=kwargs or {})
        return (msgid, r)
    except Exception as e:
        logger.exception(e)
        return (msgid, str(e))

    # Signal that we're done with this job
    return (msgid, '')



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
import time
import unittest

import mock

from .. import conf, constants, jobmanager

ADDR = 'inproc://pour_the_rice_in_the_thing'


class TestCase(unittest.TestCase):
    def test__setup(self):
        jm = jobmanager.JobManager(name='RuckusBringer')
        self.assertEqual(jm.name, 'RuckusBringer')

        self.assertFalse(jm.awaiting_startup_ack)
        self.assertEqual(jm.status, constants.STATUS.ready)

# EMQP Tests
    def test_reset(self):
        jm = jobmanager.JobManager()

        self.assertFalse(jm.awaiting_startup_ack)
        self.assertEqual(jm.status, constants.STATUS.ready)

    @mock.patch('eventmq.jobmanager.sendmsg')
    def test_send_ready(self, sndmsg_mock):
        jm = jobmanager.JobManager()
        jm.send_ready()

        sndmsg_mock.assert_called_with(jm.outgoing, 'READY')

    @mock.patch('multiprocessing.pool.Pool.close')
    @mock.patch('eventmq.jobmanager.JobManager.process_message')
    @mock.patch('eventmq.jobmanager.Sender.recv_multipart')
    @mock.patch('eventmq.jobmanager.Poller.poll')
    @mock.patch('eventmq.jobmanager.JobManager.maybe_send_heartbeat')
    @mock.patch('eventmq.jobmanager.JobManager.send_ready')
    def test__start_event_loop(self, send_ready_mock, maybe_send_hb_mock,
                               poll_mock, sender_mock, process_msg_mock,
                               pool_close_mock):
        jm = jobmanager.JobManager()
        maybe_send_hb_mock.return_value = False
        poll_mock.return_value = {jm.outgoing: jobmanager.POLLIN}
        sender_mock.return_value = [1, 2, 3]

        jm._start_event_loop()

        # send int(conf.CONCURRENT_JOBS) ready messages
        self.assertEqual(conf.CONCURRENT_JOBS, send_ready_mock.call_count)

        process_msg_mock.assert_called_with(
            sender_mock.return_value)

        jm.received_disconnect = True
        jm._start_event_loop()
        self.assertTrue(pool_close_mock.called)

    @mock.patch('eventmq.jobmanager.worker.run')
    @mock.patch('multiprocessing.pool.Pool.apply_async')
    def test_on_request(self, apply_async_mock, run_mock):
        _msgid = 'aaa0j8-ac40jf0-04tjv'
        _msg = ['a', 'b', '["run", {"a": 1}]']

        jm = jobmanager.JobManager()

        jm.on_request(_msgid, _msg)
        apply_async_mock.assert_called_with(
            args=({'a': 1}, _msgid, None),
            callback=jm.worker_done,
            func=run_mock)

    @mock.patch('eventmq.jobmanager.worker.run')
    @mock.patch('multiprocessing.pool.Pool.apply_async')
    def test_on_request_with_timeout(self, apply_async_mock, run_mock):
        timeout = 3
        _msgid = 'aaa0j8-ac40jf0-04tjv'
        _msg = ['a', 'timeout:{}'.format(timeout), '["run", {"a": 1}]']

        jm = jobmanager.JobManager()

        jm.on_request(_msgid, _msg)
        apply_async_mock.assert_called_with(
            args=({'a': 1}, _msgid, timeout),
            callback=jm.worker_done,
            func=run_mock)

    @mock.patch('eventmq.jobmanager.sendmsg')
    @mock.patch('zmq.Socket.unbind')
    def test_on_disconnect(self, socket_mock, sendmsg_mock):
        msgid = 'goog8l-uitty40-007b'
        msg = ['a', 'b', 'whatever']

        socket_mock.return_value = True

        jm = jobmanager.JobManager()
        jm.outgoing.status = constants.STATUS.listening
        jm.on_disconnect(msgid, msg)
        self.assertTrue(jm.received_disconnect, "Did not receive disconnect.")

    # Other Tests
    @mock.patch('eventmq.jobmanager.JobManager.start')
    @mock.patch('eventmq.jobmanager.import_settings')
    @mock.patch('eventmq.jobmanager.Sender.rebuild')
    def test_sighup_handler(self, rebuild_mock, import_settings_mock,
                            start_mock):
        jm = jobmanager.JobManager()

        jm.sighup_handler(982374, "FRAMEY the frame")

        self.assertTrue(rebuild_mock.called)

        # called once for the default settings, once for the jobmanager
        # settings
        self.assertEqual(2, import_settings_mock.call_count)
        # check to see if the last call was called with the jobmanager section
        import_settings_mock.assert_called_with(section='jobmanager')

        start_mock.assert_called_with(
            addr=conf.WORKER_ADDR,
        )

    @mock.patch('eventmq.jobmanager.sendmsg')
    def test_sigterm_handler(self, sendmsg_mock):
        jm = jobmanager.JobManager()

        jm.sigterm_handler(13231, "FRAMEY the evil frame")

        sendmsg_mock.assert_called_with(jm.outgoing, constants.KBYE)
        self.assertFalse(jm.awaiting_startup_ack)
        self.assertTrue(jm.received_disconnect)

    @mock.patch('eventmq.jobmanager.JobManager.start')
    @mock.patch('eventmq.jobmanager.import_settings')
    def test_jobmanager_main(self, import_settings_mock, start_mock):
        jm = jobmanager.JobManager()

        jm.jobmanager_main()

        self.assertEqual(2, import_settings_mock.call_count)
        # Assert that the last call to import settings was for the jobmanager
        # section
        import_settings_mock.assert_called_with(section='jobmanager')

        start_mock.assert_called_with(addr=conf.WORKER_ADDR,
                                      queues=conf.QUEUES)

        jm.queues = ((10, 'derp'), (0, 'blurp'))
        jm.jobmanager_main()

        start_mock.assert_called_with(addr=conf.WORKER_ADDR,
                                      queues=jm.queues)

    def cleanup(self):
        self.jm.on_disconnect(None, None)
        self.jm = None


def call_done(jm):
    jm.active_jobs += 1
    return True


def start_jm(jm, addr):
    jm.start(addr)


def pretend_job():
    time.sleep(1)



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
import unittest

import mock
from testfixtures import LogCapture

from .. import conf
from ..client import messages


class TestClass(object):
    """
    class to use in the name_from_callable test
    """
    def mymethod(self):
        """
        this method is used as the callable in
        :meth:`TestCase.test_name_from_callable`
        """
        return True


class CallableTestClass(object):
    def __call__(self):
        return True


class TestCase(unittest.TestCase):
    @mock.patch('eventmq.client.messages.send_request')
    def test_defer_job(self, sndreq_mock):
        import urlparse

        _msgid = 'mv029aisjf-09asdfualksd-aklds290fjoiw'

        sndreq_mock.return_value = _msgid
        socket = mock.Mock()

        msgid = messages.defer_job(socket, urlparse.urlsplit,
                                   args=[1, 2],
                                   kwargs={'a': 1, 'b': 2},
                                   class_args=[9, 8],
                                   class_kwargs={'z': 9, 'y': 8},
                                   reply_requested=True,
                                   guarantee=False,
                                   retry_count=3,
                                   timeout=0,
                                   queue='test_queue')
        # defer_job should return _msgid untouched
        self.assertEqual(msgid, _msgid)

        msg = ['run', {
            'callable': 'urlsplit',
            'path': 'urlparse',
            'args': [1, 2],
            'kwargs': {'a': 1, 'b': 2},
            'class_args': [9, 8],
            'class_kwargs': {'z': 9, 'y': 8},
        }]

        sndreq_mock.assert_called_with(socket, msg,
                                       reply_requested=True,
                                       guarantee=False,
                                       retry_count=3,
                                       timeout=0,
                                       queue='test_queue')

        with LogCapture() as log_checker:
            # don't blow up on an invalid callable path
            messages.defer_job(socket, 'non-callable')

            # don't blow up on some random nameless func
            def nameless_func():
                return True
            nameless_func.func_name = ''
            messages.defer_job(socket, nameless_func)

            # don't blow-up for pathless functions
            nameless_func.func_name = 'nameless_func'
            nameless_func.__module__ = None
            messages.defer_job(socket, nameless_func)

            # log an error if a callable instance object is passed
            callable_obj = CallableTestClass()
            messages.defer_job(socket, callable_obj)

            log_checker.check(
                ('eventmq.client.messages',
                 'ERROR',
                 'Invalid callable string passed, absolute path required: "non-callable"'),  # noqa
                ('eventmq.utils.functions',
                 'ERROR',
                 'Encountered callable with no name in eventmq.tests.test_client_messages'),  # noqa
                ('eventmq.client.messages',
                 'ERROR',
                 'Encountered invalid callable, will not proceed.'),
                ('eventmq.utils.functions',
                 'ERROR',
                 'Encountered callable with no __module__ path nameless_func'),
                ('eventmq.client.messages',
                 'ERROR',
                 'Encountered invalid callable, will not proceed.'),
                ('eventmq.utils.functions',
                 'ERROR',
                 'Encountered unknown callable ({}) type instanceobject'.format(callable_obj)),  # noqa
                ('eventmq.client.messages',
                 'ERROR',
                 'Encountered invalid callable, will not proceed.'),
            )

    def test_name_from_callable(self):
        import mimetools
        funcpath = messages.name_from_callable(mimetools.choose_boundary)

        t = TestClass()
        methpath = messages.name_from_callable(t.mymethod)

        self.assertEqual(funcpath, 'mimetools.choose_boundary')
        self.assertEqual(
            methpath,
            'eventmq.tests.test_client_messages:TestClass.mymethod')

    @mock.patch('eventmq.client.messages.send_schedule_request')
    def test_schedule(self, send_schedule_req_mock):
        import mimetools
        _msgid = 'ovznopi4-)*(@#$Nn0av84-a0cn84n03'
        send_schedule_req_mock.return_value = _msgid

        socket = mock.Mock()

        msgid = messages.schedule(socket, mimetools.decode,
                                  interval_secs=500,
                                  args=(1, 2),
                                  kwargs={'a': 1, 'b': 2},
                                  class_args=(9, 8),
                                  class_kwargs={'z': 9, 'y': 8},
                                  headers=('guarantee', 'poop'),
                                  queue='blurp')

        # send_schedule should return an untouched message id
        self.assertEqual(msgid, _msgid)

        msg = ['run', {
            'callable': 'decode',
            'path': 'mimetools',
            'args': (1, 2),
            'kwargs': {'a': 1, 'b': 2},
            'class_args': (9, 8),
            'class_kwargs': {'z': 9, 'y': 8},
        }]

        send_schedule_req_mock.assert_called_with(
            socket,
            interval_secs=500,
            cron='',  # default arg
            message=msg,
            headers=('guarantee', 'poop'),
            queue='blurp',
            unschedule=False)  # default arg

        with LogCapture() as log_checker:
            # don't blow up on an invalid callable path
            messages.schedule(socket, 'non-callable',
                              class_args=(123,),
                              interval_secs=10)

            # don't blow up on some random nameless func
            def nameless_func():
                return True
            nameless_func.func_name = ''
            messages.schedule(socket, nameless_func,
                              class_args=(48,),
                              interval_secs=19)

            # don't blow-up for pathless functions
            nameless_func.func_name = 'nameless_func'
            nameless_func.__module__ = None
            messages.schedule(socket, nameless_func,
                              class_args=(123,),
                              interval_secs=8920)

            # log an error if a callable instance object is passed
            callable_obj = CallableTestClass()
            messages.schedule(socket, callable_obj,
                              class_args=(123,),
                              interval_secs=23)

            # error if class_args does't have an inital id. (if this test needs
            # to be removed, also remove the class_args from the above calls)
            messages.schedule(socket, mimetools.decode)

            # error if neither cron or interval_secs is specified
            messages.schedule(socket, mimetools.decode, class_args=(123,))

            log_checker.check(
                ('eventmq.client.messages',
                 'ERROR',
                 'Invalid callable string passed, absolute path required: "non-callable"'),  # noqa
                ('eventmq.utils.functions',
                 'ERROR',
                 'Encountered callable with no name in eventmq.tests.test_client_messages'),  # noqa
                ('eventmq.client.messages',
                 'ERROR',
                 'Encountered invalid callable, will not proceed.'),
                ('eventmq.utils.functions',
                 'ERROR',
                 'Encountered callable with no __module__ path nameless_func'),
                ('eventmq.client.messages',
                 'ERROR',
                 'Encountered invalid callable, will not proceed.'),
                ('eventmq.utils.functions',
                 'ERROR',
                 'Encountered unknown callable ({}) type instanceobject'.format(callable_obj)),  # noqa
                ('eventmq.client.messages',
                 'ERROR',
                 'Encountered invalid callable, will not proceed.'),
                ('eventmq.client.messages',
                 'ERROR',
                 'First `class_args` argument must be caller_id for scheduling interval jobs'),  # noqa
                ('eventmq.client.messages',
                 'ERROR',
                 'You must sepcify either `interval_secs` or `cron`, but not both (or neither)'),  # noqa
            )

    @mock.patch('eventmq.client.messages.send_emqp_message')
    def test_send_request(self, snd_empq_msg_mock):
        _msgid = '0svny2rj8d0-aofinsud4839'
        snd_empq_msg_mock.return_value = _msgid

        socket = mock.Mock()

        msg = {'alksjfd': [1, 2],
               'laksdjf': 4,
               'alkfjds': 'alksdjf'}

        msgid = messages.send_request(socket, msg,
                                      reply_requested=True,
                                      guarantee=False,
                                      retry_count=2,
                                      queue='mozo')
        self.assertEqual(msgid, _msgid)
        snd_empq_msg_mock.assert_called_with(
            socket, 'REQUEST',
            ('mozo',
             'reply-requested,retry-count:2',
             messages.serialize(msg)))

    @mock.patch('eventmq.client.messages.send_emqp_message')
    def test_send_request_all_headers(self, snd_empq_msg_mock):
        _msgid = '0svny2rj8d0-aofinsud4839'
        snd_empq_msg_mock.return_value = _msgid

        socket = mock.Mock()

        msg = {'alksjfd': [1, 2],
               'laksdjf': 4,
               'alkfjds': 'alksdjf'}

        msgid = messages.send_request(socket, msg,
                                      reply_requested=True,
                                      guarantee=True,
                                      retry_count=2,
                                      timeout=3)
        self.assertEqual(msgid, _msgid)
        snd_empq_msg_mock.assert_called_with(
            socket, 'REQUEST',
            ('default',
             'reply-requested,guarantee,retry-count:2,timeout:3',
             messages.serialize(msg)))

    @mock.patch('eventmq.client.messages.send_emqp_message')
    def test_send_schedule_request(self, snd_empq_msg_mock):
        _msgid = 'va08n45-lanf548afn984-m7489vs'
        snd_empq_msg_mock.return_value = _msgid

        socket = mock.Mock()

        msg = {'20if': [1, 2],
               'mu8vc': 5,
               'zhx7': {'a': 1}}

        msgid = messages.send_schedule_request(socket, msg,
                                               interval_secs=38)
        self.assertEqual(msgid, _msgid)

        snd_empq_msg_mock.assert_called_with(
            socket, 'SCHEDULE',
            (conf.DEFAULT_QUEUE_NAME,
             '',
             '38',
             messages.serialize(msg),
             '')
        )

        msgid = messages.send_schedule_request(socket, msg,
                                               interval_secs=92,
                                               unschedule=True)
        self.assertEqual(msgid, _msgid)

        snd_empq_msg_mock.assert_called_with(
            socket, 'UNSCHEDULE',
            (conf.DEFAULT_QUEUE_NAME,
             '',
             '92',
             messages.serialize(msg),
             '')
        )
